# control the MKS_SERVO42C stepper with the serial port (UART)
# the bluetooth has to be disabled on the raspi zero 2w
# requires the following lines to be written to /boot/firmware/config.txt (and reboot)
# enable_uart=1
# dtoverlay=disable-bt

import serial # note this is pyserial not the module named serial in pip
from time import sleep
import threading
import logging

PING = b'\x3a'
STOP = b'\xf7'
READ_ANGLE = b'\x36'
READ_ANGLE_ERROR = b'\x39'

DEFAULT_MICROSTEPS = 16
# number of angle ticks returned from READ_ANGLE command per revolution
ANGLE_RESOLUTION = 65535
# A speed of 1 is this many revs/sec
SPEED1_REVS = 30000.0/(DEFAULT_MICROSTEPS * 200)/60
# The maximum positive speed in revs/ per second.
MAX_SPEED = 127 * SPEED1_REVS

class MockSerial:
    def __init__(self):
        pass
    def read(self, l):
        return b'x01'*l
    def write(self, l):
        pass
    def flush(self):
        pass

class MKSSERVO42C:
    def __init__(self):
        self.port = serial.Serial ("/dev/ttyAMA0", 38400)
        # self.port = MockSerial()
        self.port.timeout = 1
        self.port.write_timeout = 1
        self.lock = threading.Lock() 

    def ping(self):
        """
        Return true if the motor is awake and enabled
        """
        with self.lock:
            self._sendSingleByteCommand(PING)
            ans = self.port.read(3)
            return len(ans) == 3 and ans[1] == 1

    def stop(self):
        """
        Command the motor to stop immediately.
        Return true if the motor replied status ok
        """
        with self.lock:
            self._sendSingleByteCommand(STOP)
            ans = self.port.read(3)
            return len(ans) == 3 and ans[1] == b'\x01'

    def runConstantSpeed(self, speed):
        """
        Command the motor to run at a constant speed in revolutions per second
        Return true if the motor replied status ok
        """
        # convert revs/sec to valid speed range (-127 +127)
        command_speed = max(-127, min(int(speed / SPEED1_REVS), 127))

        # the first bit is direction
        if command_speed > 0:
            first_bit = 128 # (line lengthening, top of spool moves towards the wall)
        else:
            first_bit = 0 # (line shortening, top of spool moves away from the wall)

        # the next 7 bits are speed
        combined = (min(abs(command_speed), 127) + first_bit).to_bytes(1, byteorder='big')

        message = b'\xe0\xf6' + combined
        message += self._calculateChecksum(message)
        with self.lock:
            self.port.write(message)
            ans = self.port.read(3)
            return len(ans) == 3 and ans[1] == b'\x01'

    def getShaftAngle(self):
        """
        Get the absolute shaft angle since boot in revolutions as a double precision float
        return (status, result)

        When spinning with negative speeds passed to runConstantSpeed, the angle reported by getMotorShaftAngle is decreasing.
        """
        with self.lock:
            self._sendSingleByteCommand(READ_ANGLE)
            valid, ans = self._readNBytesAndVerify(6) # address byte, 32 bit integer, checksum byte
            if not valid:
                return False, 0
            motor_angle = int.from_bytes(ans[1:5], byteorder='big', signed=True)
            return True, float(motor_angle) / ANGLE_RESOLUTION

    def getShaftError(self):
        """
        Get the shaft error (difference between observed shaft and the motor's internal pid loop set point)
        This can be used as a rough indication of the tension on the line.

        Returns the shaft error in degrees.
        """
        with self.lock:
            self._sendSingleByteCommand(READ_ANGLE_ERROR)
            valid, ans = self._readNBytesAndVerify(4) # address byte, 16 bit uint, checksum byte
            if not valid:
                return False, 0
            angle_error = int.from_bytes(ans[1:3], byteorder='big', signed=True)
            return True, float(angle_error) / (ANGLE_RESOLUTION/360)

    def _readNBytesAndVerify(self, n):
        ans = self.port.read(n)
        # if the first byte is not 0xe0, the motor address, its trash.
        # sometimes there is one byte of trash. Probably noise
        if len(ans) == 0:
            return False, None
        if ans[0] != 0xe0:
            logging.debug(f'first byte is not motor address, got {ans}')
            if ans[1] == 0xe0:
                one = self.port.read(1)
                ans = ans[1:] + one # it was just off by one
            else:
                logging.debug(f'where is the motor address?')
                return False, None
        rchk = int.from_bytes(self._calculateChecksum(ans[:-1]))
        msg_rchk = ans[-1]
        if msg_rchk != rchk:
            logging.debug(f'Discarded corrupted message. checksum {rchk} != {msg_rchk}, message={ans}')
            return False, None
        return True, ans

    def _sendSingleByteCommand(self, b):
        """
        b must be a bytes object with a length of 1
        """
        message = b'\xe0' + b
        message += self._calculateChecksum(message)
        self.port.write(message)
        self.port.flush()

    def _calculateChecksum(self, message):
        """
        the last (least signifigant) byte in the sum of all the bytes in the message
        """
        return (sum(message) & 255).to_bytes(1, byteorder='big')

    def getMaxSpeed(self):
        return MAX_SPEED

if __name__ == "__main__":
    motor = MKSSERVO42C()
    assert(motor.ping())
    for i in range(25):
        motor.runConstantSpeed(i)
        sleep(0.1)
    for i in range(25, 1, -1):
        motor.runConstantSpeed(-i)
        sleep(0.1)
    motor.stop()



import asyncio
import grpc
import numpy as np

from .robot_control_service_pb2 import (
    GetObservationRequest, GetObservationResponse,
    TakeActionRequest, TakeActionResponse,
    GetGamepadActionRequest,
    GetEpisodeControlRequest, GetEpisodeControlResponse,
    Point3D
)
from .robot_control_service_pb2_grpc import RobotControlServiceServicer, add_RobotControlServiceServicer_to_server

# prefer whichever camera seems to have the best lighting, but stick to one.
PREFERRED_ANCHOR_0 = 3
PREFERRED_ANCHOR_1 = 2

# positional argument constructor for Point3D proto message
def Point3Dp(x, y, z):
    return Point3D(x=x, y=y, z=z)

class RobotControlService(RobotControlServiceServicer):
    """
    Server for accepting a connection from lerobot to control stringman
    Meant to be run by AsyncObserver
    """

    def __init__(self, app_state_manager):
        self.ob = app_state_manager # the instance of AsyncObserver
        print("RobotControlService initialized.")

    async def GetObservation(self, request: GetObservationRequest, context) -> GetObservationResponse:
        winch = self.ob.datastore.winch_line_record.getLast()
        finger = self.ob.datastore.finger.getLast()
        imu = self.ob.datastore.imu_rotvec.getLast()[1:]
        laser = self.ob.datastore.range_record.getLast()[1:]

        # ob is the instance of AsyncObserver (observer.py)
        # pe is the instance of Positioner2 (position_estimator.py)
        # gant_pos and gant_vel attributes are the output of a kalman filter continuously updated from observations.
        gant_pos = self.ob.pe.gant_pos
        gant_vel = self.ob.pe.gant_vel

        response = GetObservationResponse(
            gantry_pos=Point3Dp(*gant_pos),
            gantry_vel=Point3Dp(*gant_vel),
            winch_line_speed=float(winch[2]), # index 2 = speed
            finger_angle=float(finger[1]), # index 1 = angle
            gripper_imu_rot=Point3Dp(*imu),
            laser_rangefinder=float(laser),
            finger_pad_voltage=float(finger[2]), # index 2 = voltage
            gripper_camera=self.ob.get_last_frame('g'),
            anchor_camera_0=self.ob.get_last_frame(PREFERRED_ANCHOR_0),
            anchor_camera_1=self.ob.get_last_frame(PREFERRED_ANCHOR_1),
        )

        return response

    async def TakeAction(self, request: TakeActionRequest, context) -> TakeActionResponse:
        gantry_vel = np.array([request.gantry_vel.x, request.gantry_vel.y, request.gantry_vel.z])
        winch = request.winch_line_speed
        finger = request.finger_angle
        print(f'TakeAction received on grpc channel gantry_goal_pos={gantry_vel} winch={winch} finger={finger}')

        # Send both commands concurrently before waiting for both.
        # If AsyncObserver clipped these values, the results will be what they were clipped to
        (winch, finger), commanded_vel = await asyncio.gather(
            self.ob.send_winch_and_finger(0, finger),
            self.ob.move_direction_speed(gantry_vel)
        )

        winch=0

        return TakeActionResponse(
            gantry_vel = Point3Dp(*commanded_vel),
            winch_line_speed = float(winch),
            finger_angle = float(finger),
        )

    async def GetGamepadAction(self, request: GetGamepadActionRequest, context) -> TakeActionResponse:
        """
        get the last action that was caused directly by the gamepad.
        This should be post filtering and bounds checking.
        """
        commanded_vel, winch, finger = self.ob.last_gp_action

        return TakeActionResponse(
            gantry_vel = Point3Dp(*commanded_vel),
            winch_line_speed = float(winch),
            finger_angle = float(finger),
        )

    async def GetEpisodeControl(self, request: GetEpisodeControlRequest, context) -> GetEpisodeControlResponse:
        return GetEpisodeControlResponse(events=self.ob.get_episode_control_events())

async def start_robot_control_server(app_state_manager, port='[::]:50051'):
    server = grpc.aio.server()
    add_RobotControlServiceServicer_to_server(RobotControlService(app_state_manager), server)
    server.add_insecure_port(port)
    print(f"gRPC server listening on {port}")
    await server.start()
    return server # just save this and call stop on it when you want to terminate it


# script to wind up the correct length of line on on an anchor
import time
import RPi.GPIO as GPIO
import argparse

import nf_robot.common.definitions as model_constants
from nf_robot.robot.mks42c_motor import MKSSERVO42C
from nf_robot.robot.spools import SpiralCalculator

def main():

	parser = argparse.ArgumentParser(description="Stringman Component")
	parser.add_argument(
	        'anchor_type',
	        type=str,
	        default='anchor',
	        choices=['anchor', 'power_anchor'],
	        help="The type of anchor to set (choices: anchor, power_anchor)"
	    )
	args = parser.parse_args()

	# Write the file that differentiates power anchors from regular anchors
	full_diameter=model_constants.full_spool_diameter_fishing_line
	if args.anchor_type == 'power anchor':
		full_diameter=model_constants.full_spool_diameter_power_line

	with open('/opt/robot/server.conf', 'w') as f:
		f.write('args.anchor_type' + '\n')

	empty_diameter=model_constants.empty_spool_diameter
	full_length=model_constants.assumed_full_line_length
	gear_ratio=20/51
	sc = SpiralCalculator(empty_diameter, full_diameter, full_length, gear_ratio, -1)

	motor = MKSSERVO42C()
	assert(motor.ping())

	SWITCH_PIN = 18
	GPIO.setmode(GPIO.BCM)
	GPIO.setup(SWITCH_PIN, GPIO.IN, pull_up_down=GPIO.PUD_UP)

	print("Click switch to begin winding spool")
	count = 0
	tight = False
	while count < 50 and not tight:
	    time.sleep(0.1)
	    tight = GPIO.input(SWITCH_PIN) == 0
	assert tight, 'switch never registered any click'
	print('switch functioning normally, begin winding')

	try:
		_, revs = motor.getShaftAngle()
		finish_revs = sc.calc_za_from_length(0, revs)
		while revs > finish_revs:
			print(f'revs={revs} finish={finish_revs}')
			motor.runConstantSpeed(-6)
			time.sleep(0.2)
			_, revs = motor.getShaftAngle()
	finally:
		motor.runConstantSpeed(0)

	print('test complete. camera not checked.')

if __name__ == "__main__":
    main()


import asyncio
import numpy as np
import time

from nf_robot.generated.nf.telemetry import VidStats

class StatCounter:
    def __init__(self, ob):
        self.ob = ob
        self.detection_count = 0
        self.pending_frames_in_pool = 0
        self.latency = []
        self.framerate = []
        self.last_update = time.time()
        self.run = True
        self.mean_latency = 0

    async def stat_main(self):
        while self.run:
            now = time.time()
            elapsed = now-self.last_update
            if len(self.latency) > 0:
                self.mean_latency = np.mean(np.array(self.latency))
            mean_framerate = 0
            if len(self.framerate) > 0:
                mean_framerate = np.mean(np.array(self.framerate))
            detection_rate = self.detection_count / elapsed
            self.last_update = now
            self.latency = []
            self.framerate = []
            self.detection_count = 0
            self.ob.send_ui(vid_stats=VidStats(
                detection_rate=detection_rate,
                video_latency=self.mean_latency,
                video_framerate=mean_framerate,
            ))
            await asyncio.sleep(0.5)


import subprocess
import time
import logging
import atexit
import socket
import threading

logger = logging.getLogger(__name__)

class VideoStreamer:
    def __init__(self, width, height, fps=30, rtmp_url=None):
        self.rtmp_url = rtmp_url
        self.width = width
        self.height = height
        self.fps = fps # an estimate, nothing bad happens if you fail to call send_frame() exactly at this rate
        self.process = None
        self.connection_status = 'ok'
        self.local_udp_port = None
        
        # If no RTMP URL is provided, broadcast on a random free local UDP port.
        if rtmp_url is None:
            with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
                s.bind(('127.0.0.1', 0))
                self.local_udp_port = s.getsockname()[1]

        atexit.register(self.stop)

    def _calculate_bitrate(self):
        # Bitrate estimation based on a 0.5 bits-per-pixel heuristic for H.264.
        raw_bitrate = int(self.width * self.height * self.fps * 0.5)
        target_bitrate = max(200000, min(raw_bitrate, 2500000))
        return f"{target_bitrate // 1000}k"

    def start(self):
        """Starts the FFMPEG process and the stderr monitoring thread."""
        if self.process:
            return

        # A 2-second GOP (Group of Pictures) ensures that clients joining a live 
        # stream don't have to wait more than 2 seconds for a keyframe.
        gop_size = max(1, int(self.fps * 2))
        bitrate = self._calculate_bitrate()

        command = [
            'ffmpeg',
            '-y',
            '-use_wallclock_as_timestamps', '1',
            '-f', 'rawvideo',
            '-vcodec', 'rawvideo',
            '-pix_fmt', 'rgb24',
            '-s', f'{self.width}x{self.height}',
            '-i', '-',
            '-c:v', 'libx264',
            '-pix_fmt', 'yuv420p',
            '-preset', 'ultrafast',
            '-tune', 'zerolatency',
            '-g', str(gop_size),
            '-b:v', bitrate,
        ]
            
        if self.rtmp_url:
            command.extend(['-f', 'flv', self.rtmp_url])

        if self.local_udp_port:
            # mpegts over UDP for low-latency local previews.
            command.extend([
                '-f', 'mpegts', 
                f'udp://127.0.0.1:{self.local_udp_port}?pkt_size=1316'
            ])

        # stderr must be piped to monitor for connection losses.
        self.process = subprocess.Popen(
            command, 
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=0 # Use unbuffered I/O for immediate error detection
        )

        # Launching a daemon thread ensures the stream monitoring doesn't block 
        # the main process and shuts down automatically when the main thread exits.
        self.monitor_thread = threading.Thread(target=self._monitor_stderr, daemon=True)
        self.monitor_thread.start()

        logger.info(f"FFmpeg streamer started to {self.rtmp_url or 'local UDP'}")

    def _monitor_stderr(self):
        """
        Continuously reads FFmpeg's stderr to detect crashes or connection issues.
        """
        if not self.process or not self.process.stderr:
            return

        # Iterating over readline blocks until a line is available or the pipe closes.
        for line_bytes in iter(self.process.stderr.readline, b''):
            line = line_bytes.decode('utf-8', errors='ignore').strip()

            # Skip noise, but log actual errors.
            if "Error" in line or "failed" in line or "Connection" in line:
                logger.error(f"FFmpeg Error: {line}")

            # Specific triggers for reconnection logic or auth failure.
            # "Broken pipe" usually happens when the remote server stops accepting data.
            if "Broken pipe" in line or "Connection reset" in line:
                logger.warning("Media server disconnected. Updating status to error.")
                self.connection_status = 'error'
                self.stop()
                break
            
            if "not authorized" in line.lower() or "Authentication failed" in line:
                logger.error("Streaming unauthorized. Video can only be published from robots that are sending telemetry.")
                self.connection_status = 'unauthorized'
                self.stop()
                break

    def send_frame(self, frame):
        """
        Encodes and pushes a single frame to the stream.
        """
        if not self.process or self.connection_status != 'ok':
            return

        try:
            self.process.stdin.write(frame.tobytes())
            self.process.stdin.flush()
        except (BrokenPipeError, OSError) as e:
            # Write failures occur immediately if the background process died.
            self._handle_crash(e)

    def _handle_crash(self, exception):
        """Cleanup and update status when a write failure occurs."""
        logger.error(f"FFmpeg pipe broken during write: {exception}")
        self.connection_status = 'error'
        self.stop()

    def stop(self):
        """Gracefully shuts down the FFmpeg process."""
        if self.process:
            try:
                if self.process.stdin:
                    self.process.stdin.close()
                self.process.terminate()
                self.process.wait(timeout=2)
            except subprocess.TimeoutExpired:
                self.process.kill()
            except Exception:
                # If the process is already gone, don't raise during cleanup.
                pass
            finally:
                self.process = None

        if self.connection_status not in ['unauthorized', 'error']:
            self.connection_status = 'disconnected'


import numpy as np
from nf_robot.generated.nf import common

def tonp(vec: common.Vec3):
    return np.array([vec.x, vec.y, vec.z], dtype=float)

def fromnp(arr: np.ndarray):
    return common.Vec3(float(arr[0]), float(arr[1]), float(arr[2]))

def clamp(x,small,big):
    return max(min(x,big),small)

def remap(x, ilow, ihigh, olow, ohigh):
    return (x-ilow) / (ihigh-ilow) * (ohigh-olow) + olow 

def poseTupleToProto(p):
    return common.Pose(rotation=fromnp(p[0]), position=fromnp(p[1]))

def poseProtoToTuple(p):
    return (np.array([p.rotation.x, p.rotation.y, p.rotation.z]), np.array([p.position.x, p.position.y, p.position.z]))


# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: robot-config.proto
# plugin: python-betterproto2
# This file has been @generated

__all__ = (
    "Anchor",
    "CameraCalibration",
    "Gripper",
    "Resolution",
    "StringmanPilotConfig",
)

from dataclasses import dataclass

import betterproto2

from ...message_pool import default_message_pool

_COMPILER_VERSION = "0.9.0"
betterproto2.check_compiler_version(_COMPILER_VERSION)


@dataclass(eq=False, repr=False)
class Anchor(betterproto2.Message):
    """
    Configuration of an Anchor
    """

    num: "int" = betterproto2.field(1, betterproto2.TYPE_UINT32)
    """
    Index of the anchor in the system (0 to 4).
    """

    service_name: "str | None" = betterproto2.field(
        2, betterproto2.TYPE_STRING, optional=True
    )
    """
    123.cranebot-anchor-service.2ccf67bc3fc4
    """

    pose: "_common__.Pose | None" = betterproto2.field(
        3, betterproto2.TYPE_MESSAGE, optional=True
    )
    """
    Pose composed of rvec and tvec (standard OpenCV format)
    """

    address: "str | None" = betterproto2.field(
        4, betterproto2.TYPE_STRING, optional=True
    )
    """
    IP address where last seen
    """

    port: "int" = betterproto2.field(5, betterproto2.TYPE_UINT32)


default_message_pool.register_message("nf.config", "Anchor", Anchor)


@dataclass(eq=False, repr=False)
class CameraCalibration(betterproto2.Message):
    resolution: "Resolution | None" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, optional=True
    )

    intrinsic_matrix: "list[float]" = betterproto2.field(
        2, betterproto2.TYPE_DOUBLE, repeated=True
    )
    """
    Flattened 3x3 matrix (row-major)
    [fx, 0, cx, 0, fy, cy, 0, 0, 1]
    """

    distortion_coeff: "list[float]" = betterproto2.field(
        3, betterproto2.TYPE_DOUBLE, repeated=True
    )
    """
    1x5 distortion coefficients
    [k1, k2, p1, p2, k3]
    """


default_message_pool.register_message(
    "nf.config", "CameraCalibration", CameraCalibration
)


@dataclass(eq=False, repr=False)
class Gripper(betterproto2.Message):
    service_name: "str | None" = betterproto2.field(
        1, betterproto2.TYPE_STRING, optional=True
    )
    """
    service name also indicates which type it is, such as pilot gripper or arp.
    """

    address: "str | None" = betterproto2.field(
        2, betterproto2.TYPE_STRING, optional=True
    )

    port: "int" = betterproto2.field(3, betterproto2.TYPE_UINT32)

    frame_room_spin: "float" = betterproto2.field(4, betterproto2.TYPE_DOUBLE)
    """
    Rotation in radians
    """


default_message_pool.register_message("nf.config", "Gripper", Gripper)


@dataclass(eq=False, repr=False)
class Resolution(betterproto2.Message):
    width: "int" = betterproto2.field(1, betterproto2.TYPE_UINT32)

    height: "int" = betterproto2.field(2, betterproto2.TYPE_UINT32)


default_message_pool.register_message("nf.config", "Resolution", Resolution)


@dataclass(eq=False, repr=False)
class StringmanPilotConfig(betterproto2.Message):
    """
    Message representing the configuration read and written by observer
    """

    anchors: "list[Anchor]" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, repeated=True
    )

    camera_cal: "CameraCalibration | None" = betterproto2.field(
        2, betterproto2.TYPE_MESSAGE, optional=True
    )
    """
    The camera calibration values for a raspberry pi camera module 3 standard
    """

    camera_cal_wide: "CameraCalibration | None" = betterproto2.field(
        10, betterproto2.TYPE_MESSAGE, optional=True
    )
    """
    values for a raspberry pi camera module 3 wide
    """

    max_accel: "float" = betterproto2.field(3, betterproto2.TYPE_DOUBLE)
    """
    Max acceleration of anchor line in meters per second squarted
    """

    rec_mod: "int" = betterproto2.field(4, betterproto2.TYPE_INT32)
    """
    record line length every nth iteration of the main anchor loop
    """

    running_ws_delay: "float" = betterproto2.field(5, betterproto2.TYPE_DOUBLE)
    """
    running delay of main anchor loop in seconds.
    """

    gripper: "Gripper | None" = betterproto2.field(
        6, betterproto2.TYPE_MESSAGE, optional=True
    )

    preferred_cameras: "list[int]" = betterproto2.field(
        7, betterproto2.TYPE_INT32, repeated=True
    )
    """
    List of anchor nums of preferred anchor cameras.
    """

    robot_id: "str" = betterproto2.field(8, betterproto2.TYPE_STRING)

    has_been_calibrated: "bool" = betterproto2.field(9, betterproto2.TYPE_BOOL)


default_message_pool.register_message(
    "nf.config", "StringmanPilotConfig", StringmanPilotConfig
)


from .. import common as _common__



import asyncio
from asyncio.subprocess import PIPE, STDOUT
import os
import signal
import websockets
from websockets.exceptions import (
    ConnectionClosedOK,
    ConnectionClosedError,
)
import json
import threading
import zeroconf
from zeroconf.asyncio import (
    AsyncZeroconf,
)
import uuid
import socket
import time
import re
from getmac import get_mac_address
import argparse
import logging

import nf_robot.common.definitions as model_constants
from nf_robot.robot.spools import SpoolController
from nf_robot.robot.mks42c_motor import MKSSERVO42C

# using libav makes it possible to send a containerized stream with pts
# hardware h264 encoding is still used as long as resolution is below 1080
# this requires rpicam-apps (not present in lite OS image)
stream_command = [
    "/usr/bin/rpicam-vid", "-t", "0", "-n",
    "--width=1920", "--height=1080",
    "-o", "tcp://0.0.0.0:8888?listen=1",
    "--codec", "libav",
    "--libav-format", "mpegts",
    "--vflip", "--hflip",
    "--autofocus-mode", "manual",
    "--lens-position", "0.1",
    "--low-latency",
    "--bitrate", "1200kbps"
]

ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])') # https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
# the line we are looking for looks like this
#Output #0, mpegts, to 'tcp://0.0.0.0:8888?listen=1':
ready_line_re = re.compile(r"Output #0, mpegts, to 'tcp://([^:]+):(\d+)\?listen=1':")

# offset in seconds between the appearance of the ready line and the zero point of the DTS times in the stream container.
# determined experimentally by running experiments/measure_dts_zero_point.py on the rpi
dts_zero_offset = 0.719379

# values that can be overridden by the controller
default_conf = {
    # delay in seconds between updates sent on websocket during normal operation
    'RUNNING_WS_DELAY': 0.1,
    # delay in seconds between updates sent on websocket during calibration
    'CALIBRATING_WS_DELAY': 0.05
}

class RobotComponentServer:
    def __init__(self):
        self.conf = default_conf.copy()
        self.run_server = True
        # a dict of update to be flushed periodically to the websocket
        self.update = {}
        self.ws_delay = self.conf['RUNNING_WS_DELAY']
        self.rpicam_process = None
        self.zc = None # zerconf instance.
        self.mock_camera_port = None
        self.extra_tasks = []
        self.stream_command = stream_command # subclasses may override

    async def stream_measurements(self, ws):
        """
        stream line length measurements to the provided websocket connection
        as long as it exists
        """
        logging.info('start streaming measurements')
        while True:
            if self.spooler is not None:
                # add line lengths
                meas = self.spooler.popMeasurements()
                if len(meas) > 0:
                    if len(meas) > 50:
                        meas = meas[:50]
                    self.update['line_record'] = meas

            self.readOtherSensors()

            # send on websocket
            if self.update != {}:
                await ws.send(json.dumps(self.update))
            self.update = {}

            # chill
            await asyncio.sleep(self.ws_delay)

    async def stream_video(self, websocket):
        # keep rpicam-vid running until this task is cancelled by the client disconnecting
        while True:

            # make sure websocket is alive by running ping before starting rpicam-vid again.
            # if connected, this returns the connection latency.
            # if not connected, trying to get the result of the future throws a connection closed exception
            if not await websocket.ping():
                return

            if self.mock_camera_port is not None:
                # in a unit test, use mock camera. it's already running, just tell the client to connect to it
                print(f'Anchor server is configured to use mock camera on localhost:{self.mock_camera_port}')
                self.update['video_ready'] = (self.mock_camera_port, time.time())
                # normally the only other thing this task needs to do is watch the output of rpicam-vid and collect information
                # indiciating the wall time of the DTS zero point
                # this behavior is not at part of the test and the client will receive the default of now
                result = await asyncio.Future()
            else:
                try:
                    logging.info('Restarting rpi-cam_vid')
                    result = await self.run_rpicam_vid()
                    # if it stops, we'll have to wait a second or two for the OS to free the port it listens on
                    await asyncio.sleep(5)
                except FileNotFoundError:
                    # we may be running in a test. In this case stop attempting to run rpicam-vid.
                    # the client will never receive the message indicating it should connect to video.
                    logging.warning('/usr/bin/rpicam-vid does not exist on this system')
                    return
                except asyncio.CancelledError as e:
                    logging.info("Killing rpicam-vid subprocess the task is being cancelled")
                    try:
                        self.rpicam_process.kill()
                    except (ProcessLookupError, AttributeError):
                        pass
                    return await self.rpicam_process.wait()

    async def run_rpicam_vid(self):
        """
        Start the rpicam-vid stream process
        rpicam-vid listens for a single connection on 8888 and streams video to it, then terminates when the client disconnects.
        It prints a few setup lines and we need to record the time of one of them and inform the client of it.
        the client uses that to compute wall times from PTS times.
        it prints one more line after that then stops printing stuff until a few lines when the client disconnects.
        """
        self.rpicam_process = await asyncio.create_subprocess_exec(
            self.stream_command[0], *self.stream_command[1:], stdout=PIPE, stderr=STDOUT)
        # read all the lines of output
        while True:
            # during normal streaming, it is normal for this to block a long time because rpicam-vid isn't writing lines
            line = await self.rpicam_process.stdout.readline()
            if not line: # EOF.
                print('rpicam-vid exited')
                break
            line = line.decode()
            # remove color codes
            line = ansi_escape.sub('', line)
            print(line[:-1])

            # Look for the line indicating the stream is ready
            match = ready_line_re.match(line)
            if match:
                ready_wall_time = time.time()
                logging.info('rpicam-vid appears to be ready')
                await asyncio.sleep(1) # it's not ready quite yet
                # tell the websocket client to connect to the video stream. it will do so in another thread.
                self.update['video_ready'] = (8888, ready_wall_time + dts_zero_offset)
            else:
                # catch a few different kinds of errors that mean rpi-cam will have to be restarted
                # some of these can only happen after we have asked the client to try connecting to video.
                # and they don't result in rpicam-vid terminating on it's own.
                # ERROR: *** failed to allocate buffers
                # ERROR: *** failed to acquire camera
                # ERROR: *** failed to bind listen socket
                if line.startswith("ERROR: ***"):
                    logging.info('Killing rpicam-vid subprocess')
                    self.rpicam_process.kill()
                    break    
            # nothing wrong keep waiting for output lines

        # wait for the subprocess to exit, whether because we killed it, or it stopped normally
        return await self.rpicam_process.wait()

    async def run_update(self):
        logging.info('Performing Update')
        pip_subprocess = await asyncio.create_subprocess_exec(
            '/opt/robot/env/bin/pip', 'install', '--upgrade', '"nf_robot[pi]"', stdout=STDOUT, stderr=STDOUT)
        pip_subprocess.wait()
        logging.info('Self update complete. Restarting.')
        self.shutdown() # systemctl will bring us back up.


    async def read_updates_from_client(self,websocket,tg):
        while True:
            message = await websocket.recv()
            update = json.loads(message)

            if 'set_config_vars' in update:
                self.conf.update(update['set_config_vars'])
            if 'host_time' in update:
                logging.debug(f'measured latency = {time.time() - float(update["host_time"])}')
            if 'run_update' in update:
                self.run_update()
                self.extra_tasks.append(asyncio.create_task(self.run_update))

            if self.spooler is not None:
                if 'length_set' in update:
                    self.spooler.setTargetLength(update['length_set'])
                if 'aim_speed' in update:
                    self.spooler.setAimSpeed(update['aim_speed'])
                if 'jog' in update:
                    self.spooler.jogRelativeLen(float(update['jog']))
                if 'reference_length' in update:
                    self.spooler.setReferenceLength(float(update['reference_length']))
                if 'set_zero_angle' in update:
                    self.spooler.sc.set_zero_angle(float(update['set_zero_angle']))

            # defer to specific server subclass
            result = await self.processOtherUpdates(update,tg)

    async def handler(self,websocket):
        logging.info('Websocket connected')

        # This features requires Python3.11
        # The first time any of the tasks belonging to the group fails with an exception other than asyncio.CancelledError,
        # the remaining tasks in the group are cancelled.
        # For normal client disconnects either the streaming or reading task will throw a ConnectionClosedOK
        # and the taskgroup context manager will cancel the other tasks, and re-raise it in an ExceptionGroup
        # except* matches errors within an ExceptionGroup
        # If the thrown exception is not one of the type caught here, the server stops.
        try:
            async with asyncio.TaskGroup() as tg:
                read_updates = tg.create_task(self.read_updates_from_client(websocket, tg))
                stream = tg.create_task(self.stream_measurements(websocket))
                mjpeg = tg.create_task(self.stream_video(websocket))
        except* (ConnectionClosedOK, ConnectionClosedError):
            logging.info("Client disconnected")
        logging.info("All tasks in handler task group completed")
        # stop spool motors just in case some task left it running
        if self.spooler is not None:
            self.spooler.setAimSpeed(0)


    async def main(self, port=8765, name=None):
        logging.info('Starting cranebot server')
        loop = asyncio.get_running_loop()
        loop.add_signal_handler(getattr(signal, 'SIGINT'), self.shutdown)

        # used in testing when running multiple servers on the same machine
        if name is not None:
            self.service_name = name

        self.run_server = True
        asyncio.create_task(self.register_mdns_service(f"123.{self.service_name}", "_http._tcp.local.", port))

        # thread for controlling stepper motor
        if self.spooler is not None:
            self.extra_tasks.append(asyncio.create_task(asyncio.to_thread(self.spooler.trackingLoop)))

        # Call a function which subclasses implement to start tasks at startup that should remain running even if clients disconnect.
        # tasks started this way should run only while self.run_server is true
        # should return a list of any tasks it started
        self.extra_tasks.extend(self.startOtherTasks())

        async with websockets.serve(self.handler, "0.0.0.0", port):
            logging.info("Websocket server started")
            while self.run_server:
                await asyncio.sleep(0.5)
            # if those tasks finish, exiting this context will cause the server's close() method to be called.
            logging.info("Closing websocket server")


        await self.zc.async_unregister_all_services()
        logging.info("Service unregistered")
        if len(self.extra_tasks) > 0:
            result = await asyncio.gather(*self.extra_tasks)


    def shutdown(self):
        # this might get called twice
        if self.run_server:
            logging.info('\nStopping detection listener task')
            self.run_server = False
            if self.spooler is not None:
                logging.info('Stopping Spool Motor')
                self.spooler.fastStop()

    def get_wifi_ip(self):
        """Gets the Raspberry Pi's IP address on the Wi-Fi interface."""
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            ip = s.getsockname()[0]
            s.close()
            return ip
        except Exception as e:
            logging.error(f"Error getting IP address: {e}")
            # todo just wait 10 seconds and try again
            return None

    async def register_mdns_service(self, name, service_type, port, properties={}):
        """Registers an mDNS service on the network."""

        ip = "127.0.0.1" # if ip remains unchanged, we are in a unit test
        if self.zc is None:
            self.zc = AsyncZeroconf(ip_version=zeroconf.IPVersion.V4Only)
            ip = self.get_wifi_ip()

        logging.info(f'zeroconf instance advertising on {ip}')
        await asyncio.sleep(1)
        info = zeroconf.ServiceInfo(
            service_type,
            name + "." + service_type,
            port=port,
            properties=properties,
            addresses=[ip],
            server=name,
        )

        await self.zc.async_register_service(info)
        logging.info(f"Registered service: {name} ({service_type}) on port {port}")

default_anchor_conf = {
    # 0 or 1. provides a method of configuring that the switch is wired up backwards.
    # should be set to the value the pin will read when the line is pulled tight and the switch closes.
    'switch_tight_val': 0,

    # speed to reel in when the 'tighten' command is received. Meters of line per second
    'tightening_speed': -0.12,

    # if set, the switch is disabled and the line is always assumed to be tight
    'disable_switch': False,

    # Number of buffers to use when streaming mjpeg. Use as many as possible for high framerate without running out of ram
    'buffers': 3
}

try:
    import RPi.GPIO as GPIO
    gpio_ready = True
except RuntimeError:
    # we can only run that on an actual pi, not in a unit test.
    gpio_ready = False

SWITCH_PIN = 18

class RaspiAnchorServer(RobotComponentServer):
    def __init__(self, power_anchor=False, mock_motor=None):
        super().__init__()
        self.conf.update(default_anchor_conf)
        ratio = 20/51 # 20 drive gear teeth, 51 spool teeth.
        if mock_motor is not None:
            motor = mock_motor
        else:
            motor = MKSSERVO42C()
            motor.runConstantSpeed(0) # just in case
        if power_anchor:
            # A power anchor spool has a thicker line
            self.spooler = SpoolController(
                motor,
                empty_diameter=model_constants.empty_spool_diameter,
                full_diameter=model_constants.full_spool_diameter_power_line,
                full_length=model_constants.assumed_full_line_length,
                conf=self.conf, gear_ratio=ratio, tight_check_fn=self.tight_check)
        else:
            # other spools are wound with 50lb test braided fishing line with a thickness of 0.35mm
            self.spooler = SpoolController(
                motor,
                empty_diameter=model_constants.empty_spool_diameter,
                full_diameter=model_constants.full_spool_diameter_fishing_line,
                full_length=model_constants.assumed_full_line_length,
                conf=self.conf, gear_ratio=ratio, tight_check_fn=self.tight_check)
        unique = ''.join(get_mac_address().split(':'))

        if power_anchor:
            self.service_name = 'cranebot-anchor-power-service.' + unique
        else:
            self.service_name = 'cranebot-anchor-service.' + unique

        if gpio_ready:
            GPIO.setmode(GPIO.BCM)
            GPIO.setup(SWITCH_PIN, GPIO.IN, pull_up_down=GPIO.PUD_UP)

        self.report_raw = False

    def tight_check(self):
        """Return whether the line is tight according to the lever switch"""
        if (not gpio_ready) or (self.conf['disable_switch']):
            return True
        return GPIO.input(SWITCH_PIN) == self.conf['switch_tight_val']

    async def processOtherUpdates(self, updates, tg):
        if 'tighten' in updates:
            tg.create_task(self.tighten())
        if 'report_raw' in updates:
            self.report_raw = bool(updates['report_raw'])

    def readOtherSensors(self):
        # when in calibration mode send a message containing the last raw encoder angle
        if self.report_raw:
            self.update['last_raw_encoder'] = self.spooler.last_angle

    def startOtherTasks(self):
        return []

    async def tighten(self):
        """
        Pulls in the line until tight. If the line slips within 3 seconds,
        it reduces the speed by 30% and retries, up to 5 times.
        """
        max_retries = 5
        monitoring_duration_s = 3
        check_interval_s = 0.05
        
        current_speed = self.conf['tightening_speed']

        for attempt in range(1, max_retries + 1):
            # Pull in the line until the switch clicks
            while not self.tight_check():
                self.spooler.setAimSpeed(current_speed)
                await asyncio.sleep(check_interval_s)
            self.spooler.setAimSpeed(0)

            # Monitor for re-loosening over the next 3 seconds
            loosened = False
            # Calculate when the monitoring window should end
            end_time = time.monotonic() + monitoring_duration_s
            while time.monotonic() < end_time:
                if not self.tight_check():
                    loosened = True
                    break  # Exit monitoring loop immediately on slip
                await asyncio.sleep(check_interval_s)

            # Check the outcome
            if not loosened:
                print(f"Tightening successful on attempt {attempt}.")
                return # Success!

            # If it slipped, reduce speed and the loop will try again
            print(f"Line re-loosened on attempt {attempt}. Reducing speed and reeling in.")
            current_speed *= 0.7

        # If the loop finishes, all retries have failed
        self.spooler.setAimSpeed(0)
        logging.error(f"Failed to tighten line after {max_retries} attempts.")
        
if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s - %(levelname)s - %(message)s'
    )

    parser = argparse.ArgumentParser()
    parser.add_argument("--power", action="store_true",
                        help="Configures this anchor as the one which has the power line")
    args = parser.parse_args()

    ras = RaspiAnchorServer(args.power)
    asyncio.run(ras.main())



import cv2
import numpy as np
import scipy.optimize as optimize
import time
import glob
from itertools import combinations
import argparse
import logging

from nf_robot.robot.spools import SpiralCalculator
from nf_robot.common.pose_functions import *
import nf_robot.common.definitions as model_constants
from nf_robot.common.cv_common import *
from nf_robot.common.config_loader import *
from nf_robot.host.position_estimator import find_hang_point
from nf_robot.generated.nf import config

# Set up logging configuration
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

#the number of squares on the board (width and height)
board_w = 14
board_h = 9
# side length of one square in meters
board_dim = 0.075

def collect_images_locally_raspi(num_images, resolution_str):
    """
    Collects images locally on a Raspberry Pi using the picamera2 library.
    
    Args:
        num_images (int): The number of images to collect.
        resolution_str (str): The resolution as a string, e.g., "4608x2592".
    """
    try:
        from picamera2 import Picamera2
        from libcamera import Transform, controls
    except ImportError:
        logging.error("picamera2 or libcamera not found. This function is for Raspberry Pi only.")
        return
        
    width, height = map(int, resolution_str.split('x'))

    picam2 = Picamera2()
    capture_config = picam2.create_still_configuration(main={"size": (width, height), "format": "RGB888"})
    picam2.configure(capture_config)
    picam2.start()
    picam2.set_controls({"AfMode": controls.AfModeEnum.Manual, "LensPosition": 0.000001, "AfSpeed": controls.AfSpeedEnum.Fast}) 
    logging.info("Started Pi camera.")
    time.sleep(1)
    for i in range(num_images):
        time.sleep(1)
        im = picam2.capture_array()
        cv2.imwrite(f"images/cal/cap_{i}.jpg", im)
        time.sleep(1)
        logging.info(f'Collected ({i+1}/{num_images}) images.')

def collect_images_stream(address, num_images):
    """
    Connects to a video stream and collects a specified number of images.
    
    Args:
        address (str): The network address of the video stream.
        num_images (int): The number of images to collect.
    """
    logging.info(f'Connecting to {address}...')
    cap = cv2.VideoCapture(address)
    logging.debug(f'Video capture object: {cap}')
    last_cap_time = time.time()
    i = 0
    while i < num_images:
        ret, frame = cap.read()
        if not ret:
            logging.warning("Failed to capture frame from stream. Retrying...")
            return
        if time.time() > last_cap_time+1:
            fpath = f'images/cal2/cap_{i}.png'
            cv2.imwrite(fpath, frame)
            i += 1
            logging.info(f'Saved frame to {fpath}')
            last_cap_time = time.time()

def is_blurry(image, threshold=6.0):
    """
    Checks if an image is too blurry based on Laplacian variance.
    """
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY) #ensure grayscale
    laplacian_var = cv2.Laplacian(gray, cv2.CV_64F).var()
    return laplacian_var < threshold

# calibrate interactively
class CalibrationInteractive:
    def __init__(self, config_file):
        #Initializing variables
        board_n = board_w * board_h
        self.opts = []
        self.ipts = []
        self.intrinsic_matrix = np.zeros((3, 3), np.float32)
        self.distCoeffs = np.zeros((5, 1), np.float32)
        self.criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.1)
        self.config_file = config_file

        # prepare object points based on the actual dimensions of the calibration board
        # like (0,0,0), (25,0,0), (50,0,0) ....,(200,125,0)
        self.objp = np.zeros((board_n,3), np.float32)
        self.objp[:,:2] = np.mgrid[0:board_w,0:board_h].T.reshape(-1,2)
        self.objp = self.objp * board_dim
        logging.debug(f'Object points:\n{self.objp}')

        self.images_obtained = 0
        self.image_shape = None
        self.cnt = 0

    def addImage(self, image):
        #Convert to grayscale
        grey_image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
        self.image_shape = grey_image.shape[::-1]
        #Find chessboard corners
        logging.debug(f'Searching image {self.cnt}')
        self.cnt+=1
        found, corners = cv2.findChessboardCornersSB(grey_image, (board_w,board_h), cv2.CALIB_CB_EXHAUSTIVE + cv2.CALIB_CB_ACCURACY)
        # found, corners = cv2.findChessboardCorners(grey_image, (board_w,board_h), cv2.CALIB_CB_EXHAUSTIVE + cv2.CALIB_CB_NORMALIZE_IMAGE + cv2.CALIB_CB_ADAPTIVE_THRESH)

        if found == True:
            #Add the "true" checkerboard corners
            self.opts.append(self.objp)

            self.ipts.append(corners)
            self.images_obtained += 1 
            logging.info(f"Chessboards obtained: {self.images_obtained}")

            image = cv2.drawChessboardCorners(image, (14,9), corners, found)
        # this resize is only for display and should not affect calibration
        image = cv2.resize(image, (1920, 1080),  interpolation = cv2.INTER_LINEAR)
        cv2.imshow('img', image)
        cv2.waitKey(500)

    def calibrate(self):
        if self.images_obtained < 20:
            logging.error(f'Obtained {self.images_obtained} images of checkerboard. Required 20.')
            raise RuntimeError(f'Obtained {self.images_obtained} images of checkerboard. Required 20')

        logging.info('Running calibrations...')
        # ret, self.intrinsic_matrix, self.distCoeff, rvecs, tvecs = cv2.calibrateCamera(
        #     self.opts, self.ipts, self.image_shape, None, None)

        # Initialize the Matrix with the Image Center
        # This tells OpenCV: "Start assuming the lens is perfectly centered"
        w, h = self.image_shape
        self.intrinsic_matrix = np.array([
            [1000.0, 0.0,    w / 2.0], # f_x estimate, 0, c_x
            [0.0,    1000.0, h / 2.0], # 0, f_y estimate, c_y
            [0.0,    0.0,    1.0    ]
        ], dtype=np.float32)

        # Use Flags to Lock the Center
        # CALIB_USE_INTRINSIC_GUESS: Use the matrix above as the starting point
        # CALIB_FIX_PRINCIPAL_POINT: Do NOT move c_x and c_y during optimization
        flags = cv2.CALIB_USE_INTRINSIC_GUESS | cv2.CALIB_FIX_PRINCIPAL_POINT

        ret, self.intrinsic_matrix, self.distCoeff, rvecs, tvecs = cv2.calibrateCamera(
            self.opts, self.ipts, self.image_shape, 
            self.intrinsic_matrix, # Pass our initialized matrix
            None, 
            flags=flags # Pass our locking flags
        )

        #Save matrices
        logging.info(f"Camera calibration performed with image resolution: {self.image_shape[0]}x{self.image_shape[1]}.")
        logging.info('Intrinsic Matrix:')
        logging.info(str(self.intrinsic_matrix))
        logging.info('Distortion Coefficients:')
        logging.info(str(self.distCoeff))
        logging.info('Calibration complete.')

        #Calculate the total reprojection error.  The closer to zero the better.
        tot_error = 0
        for i in range(len(self.opts)):
            imgpoints2, _ = cv2.projectPoints(self.opts[i], rvecs[i], tvecs[i], self.intrinsic_matrix, self.distCoeff)
            error = cv2.norm(self.ipts[i], imgpoints2, cv2.NORM_L2)/len(imgpoints2)
            tot_error += error
        terr = tot_error/len(self.opts)
        logging.info(f"Total reprojection error: {terr}")

    def save(self): 
        logging.info(f'Saving data to {self.config_file}...')
        cfg = load_config(path=self.config_file)
        cfg.camera_cal.intrinsic_matrix = self.intrinsic_matrix.flatten().tolist()
        cfg.camera_cal.distortion_coeff = self.distCoeff.flatten().tolist()
        cfg.camera_cal.resolution.width = self.image_shape[0]
        cfg.camera_cal.resolution.height = self.image_shape[1]
        save_config(cfg, self.config_file)

# calibrate from files locally
def calibrate_from_files(config_file):
    ce = CalibrationInteractive(config_file)
    for filepath in glob.glob('images/cal/*.png'):
        logging.info(f"Analyzing {filepath}")
        image = cv2.imread(filepath)
        ce.addImage(image)
    ce.calibrate()
    ce.save()

def calibrate_from_stream(address, config_file):
    logging.info(f'Connecting to {address}...')
    cap = cv2.VideoCapture(address)
    logging.debug(f'Video capture object: {cap}')
    ce = CalibrationInteractive(config_file)
    i=0
    while ce.images_obtained < 20:
        ret, frame = cap.read()
        if ret and i%10==0:
            ce.addImage(frame)
        i+=1
    ce.calibrate()
    ce.save()

### the following functions pertain to calibration of the entire assembled robot, not just one camera ###

W_ORIGIN = 0.1 # increase this to to make origin errors more expensive
W_PLANAR = 1 # increase this to make anchor height deviations from the average plane more expensive

def multi_card_residuals(x, averages):
    """
    Computes the vector of residuals (differences) for least_squares.
    
    Returns:
        1D numpy array of floats. The optimizer minimizes sum(residuals**2).
        We return raw differences (meters), not squared errors.
    """
    anchor_poses = x.reshape((4, 2, 3))
    residuals = []
    
    # Iterate over every marker (Origin and Cal Assists)
    for marker_name, sightings in averages.items():

        # Calculate world positions for all cameras that saw this marker
        valid_sightings = []
        
        for anchor_idx, marker_pose_cams in enumerate(sightings):
            for marker_pose_cam in marker_pose_cams:
                print(f'marker_pose_cam=\n{marker_pose_cam}')
                if marker_pose_cam is None:
                    continue
                
                # Chain: Anchor -> Camera -> Marker
                pose_list = [
                    anchor_poses[anchor_idx],
                    model_constants.anchor_camera,
                    marker_pose_cam
                ]
                print(f'pose_list=\n{pose_list}')
                if marker_name == 'gantry':
                    pose_list.append(gantry_april_inv)
                pose_in_room = compose_poses(pose_list)
                
                # Extract translation (tvec is index 1)
                valid_sightings.append(pose_in_room[1])

        if not valid_sightings:
            continue
            
        projected_positions = np.array(valid_sightings)
        
        # Calculate Residuals based on marker type
        if marker_name == 'origin':
            # constraint 1: Origin must be at [0,0,0]
            # Residual = Position_Calculated - [0,0,0]
            # We add 3 residuals (dx, dy, dz) per sighting
            
            current_residuals = (projected_positions - np.zeros(3)) * W_ORIGIN
            residuals.extend(current_residuals.flatten())
            
        elif len(projected_positions) > 1:
            # constraint 2: Consistency
            # Residual = Position_Calculated - Average_Position
            
            # Calculate the centroid of where the anchors currently "think" the marker is
            centroid = np.mean(projected_positions, axis=0)
            
            # The error is the distance of each sighting from that shared centroid
            current_residuals = projected_positions - centroid
            residuals.extend(current_residuals.flatten())


        # constrain anchors to a flat plane in the z axis.
        if True:
            # Extract Z coordinates from all 4 anchors
            anchor_zs = anchor_poses[:, 1, 2]
            
            # Calculate the average Z plane
            avg_z = np.mean(anchor_zs)
            
            # Penalize deviation from the average plane
            z_residuals = (anchor_zs - avg_z) * W_PLANAR
            residuals.extend(z_residuals)

    return np.array(residuals)

def optimize_anchor_poses(averages):
    """Finds optimal anchor poses
    Args:
        averages: dict keyed by marker names. each value is a list of four lists of poses corresponding to the four anchors.
            poses represent the pose of the marker in the camera's referene frame.
    """

    # the detection of the origin marker alone is not sufficient to constrain anchor poses with enough accuracy.
    # using the average pose of each marker from each camera,
    # minimize the error between the room position of markers from different perspectives.
    # while also minimizing the error between the origin marker room positions and [0,0,0]

    initial_guesses = []
    # We rely on 'origin' being fully populated to seed the initial guess.
    origin_sightings = averages['origin']
    
    for i in range(4):
        origin_marker_pose = origin_sightings[i][0]
        
        # Calculate anchor pose relative to room by inverting the chain from origin
        # Anchor = Inverse(Origin_in_Cam + Room_relative_to_Origin + Cam_offset)
        guess = invert_pose(compose_poses([
            model_constants.anchor_camera,
            origin_marker_pose,
        ]))
        
        initial_guesses.append(guess)

    # 'lm' (Levenberg-Marquardt) is standard for unconstrained least squares
    print('running least squares optimization')
    result = optimize.least_squares(
        multi_card_residuals,
        np.array(initial_guesses).flatten(),
        args=(averages,),
        method='lm', 
        # max_nfev=1000,
        verbose=0
    )

    if not result.success:
        logging.error(f"Optimization failed. Status: {result.status}, Msg: {result.message}")
        return None

    # Reshape back to (4 anchors, 2 vectors, 3 coords)
    poses = result.x.reshape((4, 2, 3))
    return poses

def main():
    parser = argparse.ArgumentParser(description='Run robot calibration functions. Use --help for more details on each command.')
    parser.add_argument('--mode', type=str, choices=[
        'collect-images-stream',
        'calibrate-from-files',
        'collect-images-locally-raspi',
        'calibrate-from-stream'
    ], required=True, help='Choose the calibration function to run:\n \
            "collect-images-stream" to capture a specified number of images from a network stream; \
            "calibrate-from-files" to run camera calibration on a local set of images; \
            "collect-images-locally-raspi" to capture a specified number of images from a connected camera on a Raspberry Pi; \
            "calibrate-from-stream" to run camera calibration directly from a network stream until 20 images are collected.')
    parser.add_argument('--address', type=str, default='tcp://192.168.1.151:8888',
                        help='The network address for the video stream (used with stream modes).')
    parser.add_argument('--num-images', type=int, default=50,
                        help='The number of images to collect when using "collect-images-locally-raspi" or "collect-images-stream".')
    parser.add_argument('--resolution', type=str, default='4608x2592',
                        help='The resolution for the camera on the Raspberry Pi (e.g., "4608x2592"). Used with "collect-images-locally-raspi" mode.')
    parser.add_argument('--config', type=str, default=DEFAULT_CONFIG_PATH,
                        help='Path of the config file to write/update with calibrated values.')

    args = parser.parse_args()

    if args.mode == 'collect-images-locally-raspi':
        collect_images_locally_raspi(args.num_images, args.resolution)
    elif args.mode == 'collect-images-stream':
        collect_images_stream(args.address, args.num_images)
    elif args.mode == 'calibrate-from-files':
        calibrate_from_files(args.config)
    elif args.mode == 'calibrate-from-stream':
        calibrate_from_stream(args.address, args.config)

if __name__ == "__main__":
    main()




import asyncio
import os
import signal
import websockets
import time
import json
import cv2
import av
import numpy as np
from functools import partial
import threading
import os
from collections import defaultdict, deque
from websockets.exceptions import ConnectionClosedError, InvalidURI, InvalidHandshake, ConnectionClosedOK
import copy
from scipy.spatial.transform import Rotation

from nf_robot.common.cv_common import *
from nf_robot.common.pose_functions  import *
from nf_robot.common.util import *
import nf_robot.common.definitions as model_constants
from nf_robot.ml.target_heatmap import HM_IMAGE_RES
from nf_robot.generated.nf import telemetry, common
from nf_robot.host.video_streamer import VideoStreamer

os.environ['OPENCV_FFMPEG_CAPTURE_OPTIONS'] = 'fast;1|fflags;nobuffer|flags;low_delay'

# number of origin detections to average
max_origin_detections = 12

# fastSAM parameters
# seconds between processing frames with fastSAM. there is no need need to run it on every frame, since 
# we are looking at a relatively static image.
sam_rate = 1.0 # per second
sam_confidence_cutoff = 0.75

# the genertic client for a raspberri pi based robot component
class ComponentClient:
    def __init__(self, address, port, datastore, ob, pool, stat, telemetry_env):
        self.address = address
        self.port = port
        self.origin_poses = defaultdict(lambda: deque(maxlen=max_origin_detections))
        self.datastore = datastore
        self.ob = ob # instance of observer. mocks only need the update_avg_named_pos and send_ui methods
        self.websocket = None
        self.connected = False  # status of connection to websocket
        self.receive_task = None  # Task for receiving messages from websocket
        self.video_task = None  # Task for streaming video
        self.stream_start_ts = None
        self.pool = pool
        self.stat = stat
        self.last_gantry_frame_coords = None
        self.ct = None # task to connect to websocket
        self.save_raw = False
        self.connection_established_event = None
        self.frame = None # last frame of video seen
        self.last_frame_cap_time = None
        self.heartbeat_receipt = asyncio.Event()
        self.safety_task = None
        self.local_udp_port = None
        self.telemetry_env = telemetry_env

        # things used by jpeg/resizing thread
        self.frame_lock = threading.Lock()
        # This condition variable signals the worker when a new frame is ready
        self.new_frame_condition = threading.Condition(self.frame_lock)
        self.last_frame_resized = None
        # The final, encoded bytes for lerobot. Atomic write, so no lock needed.
        self.lerobot_jpeg_bytes = None
        self.lerobot_mode = False # when false disables constant encoded to improve performance.
        self.calibrating_room_spin = False # set to true momentarily during auto calibration

        self.config = ob.config

        self.conn_status = None # subclass needs to set this in init

    def send_conn_status(self):
        self.ob.send_ui(component_conn_status=copy.deepcopy(self.conn_status))

    def receive_video(self, port):
        video_uri = f'tcp://{self.address}:{port}'
        print(f'Connecting to {video_uri}')
        self.conn_status.video_status = telemetry.ConnStatus.CONNECTING
        # cannot send here, not in event loop
        self.notify_video = True

        options = {
            'rtsp_transport': 'tcp',
            'fflags': 'nobuffer',
            'flags': 'low_delay',
            'fast': '1',
        }

        try:

            # attempt = 3 # sometimes it just takes longer than expected.
            # while attempt>0:
            #     try:
            container = av.open(video_uri, options=options, mode='r')
            # except av.error.ConnectionRefusedError:
            #     attempt-=1
            #     if attempt==0:
            #         raise
            #     time.sleep(1.5)

            stream = next(s for s in container.streams if s.type == 'video')
            stream.thread_type = "SLICE"

            # start thread for frame risize and forwarding
            encoder_thread = None
            components_to_stream = [None, *self.config.preferred_cameras]
            if self.anchor_num in components_to_stream:
                encoder_thread = threading.Thread(target=self.frame_resizer_loop, kwargs={"cam_num": components_to_stream.index(self.anchor_num)}, daemon=True)
                encoder_thread.start()

            print(f'video connection successful')
            self.conn_status.video_status = telemetry.ConnStatus.CONNECTED
            self.notify_video = True
            lastSam = time.time()
            last_time = time.time()

            def error_callback_func(error):
                print(f"Error in pool worker: {error}")

            for av_frame in container.decode(stream):
                if not self.connected:
                    break
                # determine the wall time when the frame was captured
                timestamp = self.stream_start_ts + av_frame.time
                self.last_frame_cap_time = timestamp

                fr = av_frame.to_ndarray(format='rgb24')
                with self.new_frame_condition:
                    self.frame = fr
                    self.new_frame_condition.notify()

                # save information about stream latency and framerate
                now = time.time()
                self.stat.latency.append(now - timestamp)
                fr = 1/(now - last_time)
                self.stat.framerate.append(fr)
                last_time = now

                # send frame to apriltag detector
                if self.anchor_num is not None:
                    try:
                        if self.stat.pending_frames_in_pool < 60:
                            self.stat.pending_frames_in_pool += 1
                            self.pool.apply_async(
                                locate_markers,
                                (self.frame, self.config.camera_cal),
                                callback=partial(self.handle_detections, timestamp=timestamp),
                                error_callback=error_callback_func)
                        else:
                            pass
                            # print(f'Dropping frame because there are already too many pending.')
                            # TODO record fraction of frames which are dropped in stat collector
                    except ValueError:
                        break # the pool is not running

                # sleep is mandatory or this thread could prevent self.handle_detections from running and fill up the pool with work.
                # handle_detections runs in this process, but in a thread managed by the pool.
                time.sleep(0.005)

            if encoder_thread is not None:
                encoder_thread.join()

        except (av.error.TimeoutError, av.error.ConnectionRefusedError):
            print('no video stream available')
            self.conn_status.video_status = telemetry.ConnStatus.NOT_DETECTED
            self.notify_video = True
            return

        finally:
            if 'container' in locals():
                container.close()

    def frame_resizer_loop(self, cam_num):
        """
        This runs in a dedicated thread. It waits for a signal that a new
        frame is available, resizes it, and stabilizes it in the gripper case.
        The the frame is written to an ffmpeg subprocess that is sending the video
        to the ui over UDP or RTMP depending on where it is.

        The purpose of this method is to have a frame ready to send as fast as possible,
        As well as to present resized frames for inference networks to use.

        For the sake of performance, the UIs are made to consume a resolution identcal to the models,
        but if they needed to be different, we could just to two different resize ops.

        Numpy functions such as those used by cv2.resize actually release the GIL
        which is why this is a thread not a task (main loop can run faster this way)

        cam_num identifies which of the preferred cameras this is. 0 is the gripper, 1 and 2 are the two overhead cams.
        """

        # TODO allow these to change when in a teleop mode
        if self.anchor_num is None:
            final_shape = SF_TARGET_SHAPE # resize for centering network input
            final_fps = 10
        else:   
            final_shape = HM_IMAGE_RES # resize for target heatmap network input
            final_fps = 10

        path = f'stringman/{self.config.robot_id}/{cam_num}'

        if self.telemetry_env is None:
            rtmp = None # when in LAN mode do not upload ANYTHING to the cloud.
        elif self.telemetry_env == "local":
            rtmp = f'rtmp://localhost:1935/{path}?user=user&pass=pass'
        elif self.telemetry_env == "staging":
            rtmp = f'rtmp://media.neufangled.com:1935/{path}?user=user&pass=pass'
        elif self.telemetry_env == "production":
            rtmp = f'rtmp://media.neufangled.com:1935/{path}?user=user&pass=pass'
        else:
            rtmp = None

        # this is basically an ffmpeg subprocess
        vs = VideoStreamer(width=final_shape[0], height=final_shape[1], fps=final_fps, rtmp_url=rtmp)
        vs.start()

        frames_sent = 0
        time_last_frame_taken = time.time()-1

        while self.connected:
            with self.new_frame_condition:
                # Wait until the main receive_video loop signals us.
                # The 'wait' call will timeout after 1 second to re-check
                # the self.connected flag, allowing the thread to exit gracefully.
                signaled = self.new_frame_condition.wait(timeout=1.0)
                if not signaled:
                    continue
                # only take every nth frame based on framerate target
                now = time.time()
                if now < (time_last_frame_taken + 1/final_fps):
                    continue
                time_last_frame_taken = now
                # We were woken up, so copy the frame pointer while we have the lock
                frame_to_encode = self.frame

            if frame_to_encode is None:
                print(f'no frame to encode {self}')
                continue

            # Do the actual work outside the lock
            # This lets the receive_video loop add the next frame without waiting for the encode.
            self.last_frame_resized = self.process_frame(frame_to_encode)

            # send self.last_frame_resized to the UI process
            vs.send_frame(self.last_frame_resized)
            frames_sent += 1
            if frames_sent == 20:
                # sending the notification on the 20th frame ensures that the mediamtx server has something to send before clients connect
                self.local_udp_port = vs.local_udp_port
                self.ob.send_ui(video_ready=telemetry.VideoReady(
                    is_gripper=self.anchor_num is None,
                    anchor_num=self.anchor_num,
                    local_uri=f'udp://127.0.0.1:{vs.local_udp_port}',
                    stream_path=path,
                ))

        vs.stop()

    def process_frame(self, frame_to_encode):
        """
        Subclasses should override this function to resize or stabilize frames based on specific hardware constants
        the returned frame will be what is used for inference and sent to any teleoperation pipelines.
        Runs in a seperate thread from the main client.
        """
        return frame_to_encode

    async def connect_websocket(self):
        # main client loop
        self.conn_status.websocket_status = telemetry.ConnStatus.CONNECTING
        self.conn_status.video_status = telemetry.ConnStatus.NOT_DETECTED
        self.conn_status.ip_address = self.address
        self.send_conn_status()

        self.abnormal_shutdown = False # indicating we had a connection and then lost it unexpectedly
        self.failed_to_connect = False # indicating we failed to ever make a connection
        ws_uri = f"ws://{self.address}:{self.port}"
        print(f"Connecting to {ws_uri}...")
        try:
            async with websockets.connect(ws_uri, max_size=None, open_timeout=10) as websocket:
                self.connected = True
                print(f"Connected to {ws_uri}.")
                # Set an event that the observer is waiting on.
                if self.connection_established_event is not None:
                    self.connection_established_event.set()
                await self.receive_loop(websocket)
        except (asyncio.exceptions.CancelledError, websockets.exceptions.ConnectionClosedOK):
            pass # normal close
        except websockets.exceptions.ConnectionClosedError as e:
            print(f"Component server anum={self.anchor_num} disconnected abnormally: {e}")
            self.abnormal_shutdown = True
        except (OSError, TimeoutError, InvalidURI, InvalidHandshake) as e:
            print(f"Component server anum={self.anchor_num}: {e}")
            self.failed_to_connect = True
        finally:
            self.connected = False
        self.conn_status.websocket_status = telemetry.ConnStatus.NOT_DETECTED
        self.conn_status.video_status = telemetry.ConnStatus.NOT_DETECTED
        self.send_conn_status()
        return self.abnormal_shutdown

    async def receive_loop(self, websocket):
        self.conn_status.websocket_status = telemetry.ConnStatus.CONNECTED
        self.send_conn_status()
        # loop of a single websocket connection.
        # save a reference to this for send_commands
        self.websocket = websocket
        self.notify_video = False
        # send configuration to robot component to override default.
        r = await self.send_config()
        # start task to watch heartbeat event
        self.safety_task = asyncio.create_task(self.safety_monitor())
        vid_thread = None
        # Loop until disconnected
        while self.connected:
            try:
                message = await websocket.recv()
                # print(f'received message of length {len(message)}')
                update = json.loads(message)
                if 'video_ready' in update:
                    port = int(update['video_ready'][0])
                    self.stream_start_ts = float(update['video_ready'][1])
                    print(f'stream_start_ts={self.stream_start_ts} ({time.time()-self.stream_start_ts}s ago)')
                    vid_thread = threading.Thread(target=self.receive_video, kwargs={"port": port}, daemon=True)
                    vid_thread.start()
                # this event is used to detect an un-responsive state.
                self.heartbeat_receipt.set() 
                await self.handle_update_from_ws(update)

                # do this here because we seemingly can't do it in receive_video
                if self.notify_video:
                    self.send_conn_status()
                    self.notify_video = False

            except Exception as e:
                # don't catch websockets.exceptions.ConnectionClosedOK here because we want it to trip the infinite generator in websockets.connect
                # so it will stop retrying. after it has the intended effect, websockets.connect will raise it again, so we catch it in 
                # connect_websocket
                print(f"Connection to {self.address} closed. {e}")
                self.connected = False
                self.websocket = None
                # self.conn_status.websocket_status = telemetry.ConnStatus.NOT_DETECTED
                # self.conn_status.video_status = telemetry.ConnStatus.NOT_DETECTED
                # self.send_conn_status()
                raise e # TODO figure out if this causes the abnormal shutdown return value in connect_websocket like it should
                break
        if vid_thread is not None:
            # vid_thread should stop because self.connected is False
            vid_thread.join()

    async def send_commands(self, update):
        if self.connected:
            x = json.dumps(update)
            # by trying to get the result out of the future, you force any exception in the task to be raised
            # since this could be a websockets.exceptions.ConnectionClosedError it's important not to let it disappar
            result = await self.websocket.send(x)

    async def slow_stop_spool(self):
        # spool will decelerate at the rate allowed by the config file.
        # tracking mode will switch to 'speed'
        result = await self.send_commands({'aim_speed': 0})

    async def startup(self):
        self.ct = asyncio.create_task(self.connect_websocket())
        return await self.ct

    async def shutdown(self):
        if self.safety_task is not None:
            self.safety_task.cancel()
            result = await self.safety_task
        if self.connected:
            self.connected = False
            if not self.abnormal_shutdown and self.websocket:
                result = await self.websocket.close()
        elif self.ct:
            self.ct.cancel()
        print(f"Finished client {self.anchor_num} shutdown")

    def shutdown_sync(self):
        # this might get called twice
        print("\nWait for client shutdown (sync)")
        if self.connected:
            self.connected = False
            if self.websocket:
                asyncio.create_task(self.websocket.close())
        elif self.ct:
            self.ct.cancel()

    async def safety_monitor(self):
        """Notifies observer if this anchor stops sending line record updates for some time"""
        TIMEOUT=1 # seconds
        last_update = time.time()
        while self.connected:
            try:
                result = await asyncio.wait_for(self.heartbeat_receipt.wait(), TIMEOUT)
                # if you see the event within the timeout, all is well, clear it and wait again
                self.heartbeat_receipt.clear()
                last_update = time.time()
            except TimeoutError:
                print(f'No update sent from {self.anchor_num} in {TIMEOUT} seconds. it may have gone offline. sending ping')
                try:
                    pong_future = await self.websocket.ping()
                    latency = await asyncio.wait_for(pong_future, TIMEOUT)
                    # some hiccup on the server raspi made it unable to send anything for some time but it's not down.
                    print(f'Pong received in {latency}s, must have been my imagination.')
                    continue
                except (ConnectionClosedError, TimeoutError):
                    # it's no longer running, either because it lost power, or the server crashed.
                    print(f'Anchor {self.anchor_num} confirmed down. hasn\'t been seen in {time.time() - last_update} seconds.')
                    self.connected = False
                    # immediately trigger the "abnormal shutdown" return from the connect_websocket task
                    # this is how the observer is actually notified. follow the control flow by looking at `if abnormal_close:` in observer.py
                    if self.websocket and self.websocket.transport:
                        self.websocket.transport.close()
                except ConnectionClosedOK:
                    return
            except asyncio.exceptions.CancelledError:
                return

CAL_MARKERS = set(['origin', 'cal_assist_1', 'cal_assist_2', 'cal_assist_3'])
OTHER_MARKERS = set(['gamepad', 'hamper', 'trash', 'gamepad_back', 'hamper_back', 'trash_back'])

class RaspiAnchorClient(ComponentClient):
    def __init__(self, address, port, anchor_num, datastore, ob, pool, stat, telemetry_env):
        super().__init__(address, port, datastore, ob, pool, stat, telemetry_env)
        self.anchor_num = anchor_num # which anchor are we connected to
        self.conn_status = telemetry.ComponentConnStatus(
            is_gripper=False,
            anchor_num=self.anchor_num,
            websocket_status=telemetry.ConnStatus.NOT_DETECTED,
            video_status=telemetry.ConnStatus.NOT_DETECTED,
        )
        self.last_raw_encoder = None
        self.extratilt = 0
        self.raw_gant_poses = deque(maxlen=12)
        self.updatePose(poseProtoToTuple(self.config.anchors[anchor_num].pose))
        self.gantry_pos_sightings = deque(maxlen=100)
        self.gantry_pos_sightings_lock = threading.RLock()

    def updatePose(self, pose):
        self.anchor_pose = pose
        self.camera_pose = np.array(compose_poses([
            self.anchor_pose,
            model_constants.anchor_camera,
            (np.array([0,0,self.extratilt/180*np.pi], dtype=float), np.zeros(3, dtype=float)),
        ]))

    async def handle_update_from_ws(self, update):
        if 'line_record' in update:
            self.datastore.anchor_line_record[self.anchor_num].insertList(np.array(update['line_record']))

            # this is the event that is set when *any* anchor sends a line record.
            # used by the position estimator to immedately recalculate the hang point
            self.datastore.anchor_line_record_event.set()

        if 'last_raw_encoder' in update:
            self.last_raw_encoder = update['last_raw_encoder']

        if len(self.gantry_pos_sightings) > 0:
            with self.gantry_pos_sightings_lock:
                self.ob.send_ui(gantry_sightings=telemetry.GantrySightings(
                    sightings=[common.Vec3(*position) for position in self.gantry_pos_sightings]
                ))
                self.gantry_pos_sightings.clear()

    def handle_detections(self, detections, timestamp):
        """
        handle a list of aruco detections from the pool
        """
        self.stat.pending_frames_in_pool -= 1
        self.stat.detection_count += len(detections)

        for detection in detections:

            if detection['n'] in CAL_MARKERS:
                # save all the detections of the origin for later analysis
                self.origin_poses[detection['n']].append(detection['p'])

            if detection['n'] == 'gantry':
                # rotate and translate to where that object's origin would be
                # given the position and rotation of the camera that made this observation (relative to the origin)
                # store the time and that position in the appropriate measurement array in observer.
                # you have the pose of gantry_front relative to a particular anchor camera
                # convert it to a pose relative to the origin
                pose = np.array(compose_poses([
                    self.anchor_pose, # obtained from calibration
                    model_constants.anchor_camera, # constant
                    detection['p'], # the pose obtained just now
                    gantry_april_inv, # constant
                ]))
                position = pose.reshape(6)[3:]
                self.datastore.gantry_pos.insert(np.concatenate([[timestamp], [self.anchor_num], position])) # take only the position
                # print(f'Inserted gantry pose ts={timestamp}, pose={pose}')
                self.datastore.gantry_pos_event.set()

                self.last_gantry_frame_coords = detection['p'][1] # second item in pose tuple is position
                with self.gantry_pos_sightings_lock:
                    self.gantry_pos_sightings.append(position)

                if self.save_raw:
                    self.raw_gant_poses.append(detection['p'])

            if detection['n'] in OTHER_MARKERS:
                offset = model_constants.basket_offset_inv if detection['n'].endswith('back') else model_constants.basket_offset
                pose = np.array(compose_poses([
                    self.anchor_pose,
                    model_constants.anchor_camera, # constant
                    detection['p'], # the pose obtained just now
                    offset, # the named location is out in front of the tag.
                ]))
                position = pose.reshape(6)[3:]
                # save the position of this object for use in various planning tasks.
                self.ob.update_avg_named_pos(detection['n'], position)

    async def send_config(self):
        anchor_config_vars = {
            "MAX_ACCEL": self.config.max_accel,
            "REC_MOD": self.config.rec_mod,
            "RUNNING_WS_DELAY": self.config.running_ws_delay,
        }
        if len(anchor_config_vars) > 0:
            await self.websocket.send(json.dumps({'set_config_vars': anchor_config_vars}))

    def process_frame(self, frame_to_encode):
        return cv2.resize(frame_to_encode, HM_IMAGE_RES, interpolation=cv2.INTER_AREA)


# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: control.proto
# plugin: python-betterproto2
# This file has been @generated

__all__ = (
    "AddTargetFromAnchorCam",
    "CombinedMove",
    "Command",
    "CommonCommand",
    "ControlBatchUpdate",
    "ControlItem",
    "DeleteTarget",
    "EpControl",
    "GantryGoalPos",
    "JogSpool",
    "ScaleRoom",
)

from dataclasses import dataclass

import betterproto2

from ...message_pool import default_message_pool

_COMPILER_VERSION = "0.9.0"
betterproto2.check_compiler_version(_COMPILER_VERSION)


class Command(betterproto2.Enum):
    TIGHTEN_LINES = 0
    """
    Reel every line until tight (by switch)
    """

    HALF_CAL = 1
    """
    Perform fast calibration (takes about 2 seconds)
    """

    FULL_CAL = 2
    """
    Perform full calibration sequence. (requires setup of cards on the floor)
    """

    ZERO_WINCH = 3
    """
    Zero the winch line in the gripper by reeling it in until it hits the limit switch
    """

    STOP_ALL = 4
    """
    Stop all spool motion (while respecting deceleration limits)
    """

    ENABLE_LEROBOT = 5
    """
    Begin accepting connections from lerobot teleoperation or data collection tools
    """

    PICK_AND_DROP = 9
    """
    Begin performing automated pick and drop of targets in queue
    """

    PARK = 10
    """
    Drop everything and auto park on the saddle
    """

    UNPARK = 11
    """
    Unpark from the saddle and move clear of it.
    """

    GRASP = 12
    """
    Execute an automated grasp at the current position.
    """

    SUBMIT_TARGETS_TO_DATASET = 13
    """
    Take the current set of targets as complete and correct and add entries to dataset
    """

    HORIZONTAL_CHECK = 6
    """
    =====  Commands intended only for diagnostics =====

    A horizontal move paralell to the floor that measures the quality of calibration
    """

    COLLECT_GRIPPER_IMAGES = 7
    """
    Begin saving images of the gripper camera at regular intervals in a local directory
    """

    SHUTDOWN = 8
    """
    Shuts down the observer process. robot components remain on.
    """

    @classmethod
    def betterproto_value_to_renamed_proto_names(cls) -> dict[int, str]:
        return {
            0: "COMMAND_TIGHTEN_LINES",
            1: "COMMAND_HALF_CAL",
            2: "COMMAND_FULL_CAL",
            3: "COMMAND_ZERO_WINCH",
            4: "COMMAND_STOP_ALL",
            5: "COMMAND_ENABLE_LEROBOT",
            9: "COMMAND_PICK_AND_DROP",
            10: "COMMAND_PARK",
            11: "COMMAND_UNPARK",
            12: "COMMAND_GRASP",
            13: "COMMAND_SUBMIT_TARGETS_TO_DATASET",
            6: "COMMAND_HORIZONTAL_CHECK",
            7: "COMMAND_COLLECT_GRIPPER_IMAGES",
            8: "COMMAND_SHUTDOWN",
        }

    @classmethod
    def betterproto_renamed_proto_names_to_value(cls) -> dict[str, int]:
        return {
            "COMMAND_TIGHTEN_LINES": 0,
            "COMMAND_HALF_CAL": 1,
            "COMMAND_FULL_CAL": 2,
            "COMMAND_ZERO_WINCH": 3,
            "COMMAND_STOP_ALL": 4,
            "COMMAND_ENABLE_LEROBOT": 5,
            "COMMAND_PICK_AND_DROP": 9,
            "COMMAND_PARK": 10,
            "COMMAND_UNPARK": 11,
            "COMMAND_GRASP": 12,
            "COMMAND_SUBMIT_TARGETS_TO_DATASET": 13,
            "COMMAND_HORIZONTAL_CHECK": 6,
            "COMMAND_COLLECT_GRIPPER_IMAGES": 7,
            "COMMAND_SHUTDOWN": 8,
        }


@dataclass(eq=False, repr=False)
class AddTargetFromAnchorCam(betterproto2.Message):
    """
    Add a new target by projecting the provided point from the given camera
    and adds the current image and point to a dataset
    """

    anchor_num: "int" = betterproto2.field(1, betterproto2.TYPE_UINT32)

    img_norm_x: "float" = betterproto2.field(2, betterproto2.TYPE_FLOAT)

    img_norm_y: "float" = betterproto2.field(3, betterproto2.TYPE_FLOAT)

    target_id: "str | None" = betterproto2.field(
        4, betterproto2.TYPE_STRING, optional=True
    )
    """
    If an id is set, this is interpreted as a move command, to set a new position for the given item.
    """


default_message_pool.register_message(
    "nf.control", "AddTargetFromAnchorCam", AddTargetFromAnchorCam
)


@dataclass(eq=False, repr=False)
class CombinedMove(betterproto2.Message):
    """
    A movement command typically sent from a gamepad or policy
    """

    direction: "_common__.Vec3 | None" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, optional=True
    )
    """
    Direction in which to move gantry
    When unset, gantry continues moving according to it's last command
    when unset, speed is ignored
    """

    speed: "float | None" = betterproto2.field(
        2, betterproto2.TYPE_FLOAT, optional=True
    )
    """
    Speed to move gantry
    if speed is set, direction is normalized to a unit vector and multiplied by speed.
    if speed is unset, direction is assumed to be a velocity vector

    meters per second
    """

    finger: "float | None" = betterproto2.field(
        3, betterproto2.TYPE_FLOAT, optional=True
    )
    """
    Set the finger servo angle
    Range [-90,90]
    When unset, finger command will not be issued, finger continues tracking previously commanded angle.
    """

    winch: "float | None" = betterproto2.field(
        4, betterproto2.TYPE_FLOAT, optional=True
    )
    """
    Set the winch line speed
    Line lengthening speed in meters per second
    when unset, winch spool continues at previously commanded speed.
    not used in arp
    """

    wrist: "float | None" = betterproto2.field(
        5, betterproto2.TYPE_FLOAT, optional=True
    )
    """
    Set the wrist angle in degrees
    when unset, wrist command will not be issued, wrist continues holding previously commanded angle
    Not used in pilot
    """


default_message_pool.register_message("nf.control", "CombinedMove", CombinedMove)


@dataclass(eq=False, repr=False)
class CommonCommand(betterproto2.Message):
    """
    Start one of several common tasks which take no arguments.
    """

    name: "Command" = betterproto2.field(
        1, betterproto2.TYPE_ENUM, default_factory=lambda: Command(0)
    )


default_message_pool.register_message("nf.control", "CommonCommand", CommonCommand)


@dataclass(eq=False, repr=False)
class ControlBatchUpdate(betterproto2.Message):
    """
    the frame sent over the wire
    """

    robot_id: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)

    updates: "list[ControlItem]" = betterproto2.field(
        2, betterproto2.TYPE_MESSAGE, repeated=True
    )


default_message_pool.register_message(
    "nf.control", "ControlBatchUpdate", ControlBatchUpdate
)


@dataclass(eq=False, repr=False)
class ControlItem(betterproto2.Message):
    """
    A single kind of update that a UI, ursina or web, is providing to Observer
    Typically to control the robot

    Oneofs:
        - payload:
    """

    command: "CommonCommand | None" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    jog_spool: "JogSpool | None" = betterproto2.field(
        2, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    move: "CombinedMove | None" = betterproto2.field(
        3, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    gantry_goal_pos: "GantryGoalPos | None" = betterproto2.field(
        4, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    episode_control: "EpControl | None" = betterproto2.field(
        5, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    scale_room: "ScaleRoom | None" = betterproto2.field(
        6, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    add_cam_target: "AddTargetFromAnchorCam | None" = betterproto2.field(
        7, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    delete_target: "DeleteTarget | None" = betterproto2.field(
        8, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )


default_message_pool.register_message("nf.control", "ControlItem", ControlItem)


@dataclass(eq=False, repr=False)
class DeleteTarget(betterproto2.Message):
    target_id: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)


default_message_pool.register_message("nf.control", "DeleteTarget", DeleteTarget)


@dataclass(eq=False, repr=False)
class EpControl(betterproto2.Message):
    """
    Adds the named episode control events to a set
    When lerobot is connected, this is how a teleoperator can start or stop an episode during recording
    """

    events: "list[str]" = betterproto2.field(1, betterproto2.TYPE_STRING, repeated=True)


default_message_pool.register_message("nf.control", "EpControl", EpControl)


@dataclass(eq=False, repr=False)
class GantryGoalPos(betterproto2.Message):
    """
    Set the given goal position for the gantry, it will begin moving there.
    """

    pos: "_common__.Vec3 | None" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, optional=True
    )


default_message_pool.register_message("nf.control", "GantryGoalPos", GantryGoalPos)


@dataclass(eq=False, repr=False)
class JogSpool(betterproto2.Message):
    """
    Used for direct control of a single spool during installation or maintenence
    not for typical robot movement

    Oneofs:
        - var:
    """

    is_gripper: "bool" = betterproto2.field(1, betterproto2.TYPE_BOOL)

    anchor_num: "int" = betterproto2.field(2, betterproto2.TYPE_UINT32)
    """
    only set for anchors
    """

    speed: "float | None" = betterproto2.field(
        3, betterproto2.TYPE_FLOAT, optional=True, group="var"
    )
    """
    lengthen the line by this many meters per second until told to stop
    """

    offset: "float | None" = betterproto2.field(
        4, betterproto2.TYPE_FLOAT, optional=True, group="var"
    )
    """
    change the length of unspooled line by this many meters at a fixed and safe speed.
    """


default_message_pool.register_message("nf.control", "JogSpool", JogSpool)


@dataclass(eq=False, repr=False)
class ScaleRoom(betterproto2.Message):
    """
    message used to tweak calibration by scaling all the anchor positions towards or away from the origin
    """

    scale: "float" = betterproto2.field(1, betterproto2.TYPE_FLOAT)

    tiltcams: "float" = betterproto2.field(2, betterproto2.TYPE_FLOAT)


default_message_pool.register_message("nf.control", "ScaleRoom", ScaleRoom)


from .. import common as _common__



import sys
import logging
import time
import board
import busio

from nf_robot.robot.connect_wifi import ensure_connection

# todo maybe there is a better solution to this but systemctl starts us too early and some zeroconf things dont work
time.sleep(3)

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    filename='cranebot.log'
)

handler = logging.StreamHandler(stream=sys.stdout)
logger.addHandler(handler)

def handle_exception(exc_type, exc_value, exc_traceback):
    logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))

sys.excepthook = handle_exception

import asyncio

async def asyncmain():
    connected = await ensure_connection()
    if not connected:
        logging.error('Wifi connection script failed to find a network')
        quit()

    # determine component type
    try:
        i2c = busio.I2C(board.SCL, board.SDA)
        addrs = set(i2c.scan())
    except ValueError:
        addrs = set([])

    if addrs == set([0x29, 0x4b, 0x17]): # gripper
        from nf_robot.robot.gripper_server import RaspiGripperServer
        gs = RaspiGripperServer()
        r = await gs.main()

    elif set([0x48, 0x29]).issubset(addrs): # arpeggio_gripper with or without IMU
        from nf_robot.robot.gripper_arp_server import GripperArpServer
        gs = GripperArpServer()
        r = await gs.main()

    elif len(addrs) == 0:
        from nf_robot.robot.anchor_server import RaspiAnchorServer

        # to differentiate power anchor, look for file written by anchor_eval.py
        component_type = 'anchor'
        try:
            with open('server.conf', 'r') as file:
                for line in file:
                    line = line.strip()  # Remove leading/trailing whitespace
                    if not line.startswith('#') and line:  # Check if line is not a comment and is not empty
                        component_type = line
        except FileNotFoundError:
            component_type = 'anchor'

        powerline = component_type == 'power anchor' # TODO differentiate in some automatic way
        ras = RaspiAnchorServer(powerline)
        r = await ras.main()

def main():
    asyncio.run(asyncmain())

if __name__ == "__main__":
    main()


# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: telemetry.proto
# plugin: python-betterproto2
# This file has been @generated

__all__ = (
    "AnchorPoses",
    "CommandedVelocity",
    "ComponentConnStatus",
    "ConnStatus",
    "GantrySightings",
    "GripCamPredictions",
    "GripperModel",
    "GripperSensors",
    "NamedObjectPosition",
    "OneTarget",
    "Popup",
    "PositionEstimate",
    "PositionFactors",
    "TargetList",
    "TargetStatus",
    "TelemetryBatchUpdate",
    "TelemetryItem",
    "UplinkStatus",
    "VidStats",
    "VideoReady",
)

from dataclasses import dataclass

import betterproto2

from ...message_pool import default_message_pool

_COMPILER_VERSION = "0.9.0"
betterproto2.check_compiler_version(_COMPILER_VERSION)


class ConnStatus(betterproto2.Enum):
    NOT_DETECTED = 0

    CONNECTING = 1

    CONNECTED = 2

    @classmethod
    def betterproto_value_to_renamed_proto_names(cls) -> dict[int, str]:
        return {
            0: "CONNSTATUS_NOT_DETECTED",
            1: "CONNSTATUS_CONNECTING",
            2: "CONNSTATUS_CONNECTED",
        }

    @classmethod
    def betterproto_renamed_proto_names_to_value(cls) -> dict[str, int]:
        return {
            "CONNSTATUS_NOT_DETECTED": 0,
            "CONNSTATUS_CONNECTING": 1,
            "CONNSTATUS_CONNECTED": 2,
        }


class GripperModel(betterproto2.Enum):
    PILOT = 0

    ARPEGGIO = 1

    @classmethod
    def betterproto_value_to_renamed_proto_names(cls) -> dict[int, str]:
        return {
            0: "GRIPPERMODEL_PILOT",
            1: "GRIPPERMODEL_ARPEGGIO",
        }

    @classmethod
    def betterproto_renamed_proto_names_to_value(cls) -> dict[str, int]:
        return {
            "GRIPPERMODEL_PILOT": 0,
            "GRIPPERMODEL_ARPEGGIO": 1,
        }


class TargetStatus(betterproto2.Enum):
    SEEN = 0

    SELECTED = 1

    PICKED_UP = 2

    DROPPED = 3

    @classmethod
    def betterproto_value_to_renamed_proto_names(cls) -> dict[int, str]:
        return {
            0: "TARGETSTATUS_SEEN",
            1: "TARGETSTATUS_SELECTED",
            2: "TARGETSTATUS_PICKED_UP",
            3: "TARGETSTATUS_DROPPED",
        }

    @classmethod
    def betterproto_renamed_proto_names_to_value(cls) -> dict[str, int]:
        return {
            "TARGETSTATUS_SEEN": 0,
            "TARGETSTATUS_SELECTED": 1,
            "TARGETSTATUS_PICKED_UP": 2,
            "TARGETSTATUS_DROPPED": 3,
        }


@dataclass(eq=False, repr=False)
class AnchorPoses(betterproto2.Message):
    """
    Send by the observer to UIs when calibration or config file reading has set the anchor poses
    """

    poses: "list[_common__.Pose]" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, repeated=True
    )
    """
    always send all four poses in order of anchor num
    """


default_message_pool.register_message("nf.telemetry", "AnchorPoses", AnchorPoses)


@dataclass(eq=False, repr=False)
class CommandedVelocity(betterproto2.Message):
    """
    When the gantry has been commanded to move with a velocity eitehr by AI or human, it gives this velocity to the UI for visualization.
    """

    velocity: "_common__.Vec3 | None" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, optional=True
    )


default_message_pool.register_message(
    "nf.telemetry", "CommandedVelocity", CommandedVelocity
)


@dataclass(eq=False, repr=False)
class ComponentConnStatus(betterproto2.Message):
    """
    Update about the status of the connection between observer and a robot component such as an anchor or gripper
    """

    is_gripper: "bool" = betterproto2.field(1, betterproto2.TYPE_BOOL)

    anchor_num: "int" = betterproto2.field(2, betterproto2.TYPE_UINT32)
    """
    will be set only for anchors
    """

    websocket_status: "ConnStatus" = betterproto2.field(
        3, betterproto2.TYPE_ENUM, default_factory=lambda: ConnStatus(0)
    )

    video_status: "ConnStatus" = betterproto2.field(
        4, betterproto2.TYPE_ENUM, default_factory=lambda: ConnStatus(0)
    )

    ip_address: "str" = betterproto2.field(5, betterproto2.TYPE_STRING)
    """
    On the wifi network in the house where the robot is.
    """

    gripper_model: "GripperModel | None" = betterproto2.field(
        6, betterproto2.TYPE_ENUM, optional=True
    )


default_message_pool.register_message(
    "nf.telemetry", "ComponentConnStatus", ComponentConnStatus
)


@dataclass(eq=False, repr=False)
class GantrySightings(betterproto2.Message):
    """
    Individual sightings of the gantry by apriltag
    """

    sightings: "list[_common__.Vec3]" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, repeated=True
    )


default_message_pool.register_message(
    "nf.telemetry", "GantrySightings", GantrySightings
)


@dataclass(eq=False, repr=False)
class GripCamPredictions(betterproto2.Message):
    """
    values predicted by model using the gripper camera provided to UI for diagnostic visualization
    """

    move_x: "float" = betterproto2.field(1, betterproto2.TYPE_FLOAT)
    """
    prediction motion vector to center gripper over target
    """

    move_y: "float" = betterproto2.field(2, betterproto2.TYPE_FLOAT)

    prob_target_in_view: "float" = betterproto2.field(3, betterproto2.TYPE_FLOAT)
    """
    probability a target is in view
    """

    prob_holding: "float" = betterproto2.field(4, betterproto2.TYPE_FLOAT)
    """
    probability an object is held in the gripper (visual only)
    """

    grip_angle: "float" = betterproto2.field(5, betterproto2.TYPE_FLOAT)
    """
    Predicted ideal grip angle from vertical in the image [0-pi]
    """


default_message_pool.register_message(
    "nf.telemetry", "GripCamPredictions", GripCamPredictions
)


@dataclass(eq=False, repr=False)
class GripperSensors(betterproto2.Message):
    """
    Latest state of sensors in the gripper
    """

    range: "float" = betterproto2.field(1, betterproto2.TYPE_FLOAT)
    """
    measured laser range from palm to obstacle in meters. Unset when infinite
    """

    angle: "float" = betterproto2.field(2, betterproto2.TYPE_FLOAT)
    """
    Finger angle from full open to full closed [-90,90]
    In the pilot gripper this is the value last commanded of the finger servo.
    In the arpeggio gripper, this is the last measured value
    map this to [0,60] to get the angle in degrees of the first finger bone.
    the second bone is a deterministic function of the first one, but I don't know the function.
    """

    pressure: "float" = betterproto2.field(3, betterproto2.TYPE_FLOAT)
    """
    The voltage on the sensing finger pad, higher is more pressure. [0,3.3]
    """

    wrist: "float" = betterproto2.field(4, betterproto2.TYPE_FLOAT)
    """
    Wrist angle in degrees [0,360]
    """


default_message_pool.register_message("nf.telemetry", "GripperSensors", GripperSensors)


@dataclass(eq=False, repr=False)
class NamedObjectPosition(betterproto2.Message):
    """
    Position of a named object that UI might want to visually represent.
    only sent when it moves
    """

    position: "_common__.Vec3 | None" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, optional=True
    )
    """
    if position is unset, the UI should hide the indicator. 
    """

    name: "str" = betterproto2.field(2, betterproto2.TYPE_STRING)
    """
    Name of the object
    Could be the name of a tag like 'hamper' or 'gamepad'
    When the observer is moving the gantry to a particular goal position, it gives this position with the name 'gantry_goal'
    """


default_message_pool.register_message(
    "nf.telemetry", "NamedObjectPosition", NamedObjectPosition
)


@dataclass(eq=False, repr=False)
class OneTarget(betterproto2.Message):
    """


    Oneofs:
        - dropoff: intended dropoff location
    """

    id: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)
    """
    stable target identifier
    """

    position: "_common__.Vec3 | None" = betterproto2.field(
        2, betterproto2.TYPE_MESSAGE, optional=True
    )
    """
    position in world space
    """

    coords: "_common__.Vec3 | None" = betterproto2.field(
        3, betterproto2.TYPE_MESSAGE, optional=True, group="dropoff"
    )

    tag: "str | None" = betterproto2.field(
        4, betterproto2.TYPE_STRING, optional=True, group="dropoff"
    )

    status: "TargetStatus" = betterproto2.field(
        5, betterproto2.TYPE_ENUM, default_factory=lambda: TargetStatus(0)
    )
    """
    status of target
    """

    source: "str" = betterproto2.field(6, betterproto2.TYPE_STRING)
    """
    the entity that submitted the target to the observer.
    """


default_message_pool.register_message("nf.telemetry", "OneTarget", OneTarget)


@dataclass(eq=False, repr=False)
class Popup(betterproto2.Message):
    """
    A connected UI should immediately display this message in a popup with an OK button.
    """

    message: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)


default_message_pool.register_message("nf.telemetry", "Popup", Popup)


@dataclass(eq=False, repr=False)
class PositionEstimate(betterproto2.Message):
    """
    estimate of the robot's position
    """

    gantry_position: "_common__.Vec3 | None" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, optional=True
    )

    gantry_velocity: "_common__.Vec3 | None" = betterproto2.field(
        2, betterproto2.TYPE_MESSAGE, optional=True
    )

    gripper_pose: "_common__.Pose | None" = betterproto2.field(
        3, betterproto2.TYPE_MESSAGE, optional=True
    )
    """
    in world space
    """

    data_ts: "float" = betterproto2.field(4, betterproto2.TYPE_DOUBLE)
    """
    the wall time in seconds since the epoch UTC when the robot was at this position
    """

    slack: "list[bool]" = betterproto2.field(5, betterproto2.TYPE_BOOL, repeated=True)
    """
    an array of bools for each anchor. true meaning the line should be drawn with a droop
    """


default_message_pool.register_message(
    "nf.telemetry", "PositionEstimate", PositionEstimate
)


@dataclass(eq=False, repr=False)
class PositionFactors(betterproto2.Message):
    """
    Diagnostic information about factors used in estimating the robot's positon
    """

    visual_pos: "_common__.Vec3 | None" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, optional=True
    )
    """
    Visually determined (apriltags) position and velocity
    """

    visual_vel: "_common__.Vec3 | None" = betterproto2.field(
        2, betterproto2.TYPE_MESSAGE, optional=True
    )

    hanging_pos: "_common__.Vec3 | None" = betterproto2.field(
        3, betterproto2.TYPE_MESSAGE, optional=True
    )
    """
    Encoder determined position and velocity of hang point
    """

    hanging_vel: "_common__.Vec3 | None" = betterproto2.field(
        4, betterproto2.TYPE_MESSAGE, optional=True
    )


default_message_pool.register_message(
    "nf.telemetry", "PositionFactors", PositionFactors
)


@dataclass(eq=False, repr=False)
class TargetList(betterproto2.Message):
    """
    A full snapshot of the list of targets known to the observer and their statuses. replace previous
    """

    targets: "list[OneTarget]" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, repeated=True
    )
    """
    order of targets is the order in which the observer will try to pick them up
    """


default_message_pool.register_message("nf.telemetry", "TargetList", TargetList)


@dataclass(eq=False, repr=False)
class TelemetryBatchUpdate(betterproto2.Message):
    """
    the frame sent over the wire
    """

    robot_id: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)

    updates: "list[TelemetryItem]" = betterproto2.field(
        2, betterproto2.TYPE_MESSAGE, repeated=True
    )


default_message_pool.register_message(
    "nf.telemetry", "TelemetryBatchUpdate", TelemetryBatchUpdate
)


@dataclass(eq=False, repr=False)
class TelemetryItem(betterproto2.Message):
    """
    A single kind of update that the observer is providing to an external UI
    Typically about the robot's state such as it's position

    Oneofs:
        - payload:
    """

    pos_estimate: "PositionEstimate | None" = betterproto2.field(
        1, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    pos_factors_debug: "PositionFactors | None" = betterproto2.field(
        2, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    gantry_sightings: "GantrySightings | None" = betterproto2.field(
        3, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    new_anchor_poses: "AnchorPoses | None" = betterproto2.field(
        4, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    component_conn_status: "ComponentConnStatus | None" = betterproto2.field(
        5, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    vid_stats: "VidStats | None" = betterproto2.field(
        6, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    named_position: "NamedObjectPosition | None" = betterproto2.field(
        7, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    last_commanded_vel: "CommandedVelocity | None" = betterproto2.field(
        8, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    pop_message: "Popup | None" = betterproto2.field(
        9, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    grip_sensors: "GripperSensors | None" = betterproto2.field(
        10, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    grip_cam_preditions: "GripCamPredictions | None" = betterproto2.field(
        11, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )
    """
    note misspelled. not changed for backwards compat
    """

    target_list: "TargetList | None" = betterproto2.field(
        12, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    video_ready: "VideoReady | None" = betterproto2.field(
        13, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )

    uplink_status: "UplinkStatus | None" = betterproto2.field(
        15, betterproto2.TYPE_MESSAGE, optional=True, group="payload"
    )
    """
    only sent from control plane
    """

    retain_key: "str | None" = betterproto2.field(
        14, betterproto2.TYPE_STRING, optional=True
    )
    """
    If set, the Control Plane will cache this item using this string as the key.
    When a new UI connects, it receives the latest item for every unique key.
    e.g., "anchor_poses", "conn_status_anchor_0", "conn_status_gripper"
    """


default_message_pool.register_message("nf.telemetry", "TelemetryItem", TelemetryItem)


@dataclass(eq=False, repr=False)
class UplinkStatus(betterproto2.Message):
    online: "bool" = betterproto2.field(1, betterproto2.TYPE_BOOL)


default_message_pool.register_message("nf.telemetry", "UplinkStatus", UplinkStatus)


@dataclass(eq=False, repr=False)
class VideoReady(betterproto2.Message):
    """
    Indicates the availability of a video stream
    """

    is_gripper: "bool" = betterproto2.field(1, betterproto2.TYPE_BOOL)
    """
    is this the gripper camera?
    """

    anchor_num: "int | None" = betterproto2.field(
        2, betterproto2.TYPE_UINT32, optional=True
    )
    """
    Camera anchor num 
    """

    local_uri: "str | None" = betterproto2.field(
        3, betterproto2.TYPE_STRING, optional=True
    )
    """
    for local UI's connect at this address
    udp:127.0.0.1:1234
    """

    stream_path: "str | None" = betterproto2.field(
        4, betterproto2.TYPE_STRING, optional=True
    )
    """
    for remote UI's construct a url using this stream path
    for example http://localhost:8889/${streamPath}/whep
    """


default_message_pool.register_message("nf.telemetry", "VideoReady", VideoReady)


@dataclass(eq=False, repr=False)
class VidStats(betterproto2.Message):
    """
    Info on the throughput and latency of video connections and apriltag processing
    """

    detection_rate: "float" = betterproto2.field(1, betterproto2.TYPE_FLOAT)
    """
    tags detected per second summed accross all cameras
    """

    video_latency: "float" = betterproto2.field(2, betterproto2.TYPE_FLOAT)
    """
    seconds between when frames are captured and when the observer can process them.
    """

    video_framerate: "float" = betterproto2.field(3, betterproto2.TYPE_FLOAT)
    """
    average framerate across all cameras
    """


default_message_pool.register_message("nf.telemetry", "VidStats", VidStats)


from .. import common as _common__



import asyncio
import subprocess

REMOTE_HOSTS = [
    "nathan@192.168.1.151",
    "nathan@192.168.1.152",
    "nathan@192.168.1.153",
    "nathan@192.168.1.154",
    "nathan@192.168.1.156",
]
FILES = [
    "spools.py",
    "motor_control.py",
    "anchor_server.py",
    "gripper_server.py",
]
REMOTE_DIR = "cranebot3-firmware"

# blocking version
def run_command(command):
    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    if process.returncode != 0:
        print(f"Error executing command: {command}")
        print(stderr.decode())
        return False
    return True

async def copy_up(host):
    print(f"Copying files to {host}:{REMOTE_DIR}...")
    run_command(f"scp {" ".join(FILES)} {host}:{REMOTE_DIR}")
    # process = await asyncio.create_subprocess_shell(
    #     f"scp {" ".join(FILES)} {host}:{REMOTE_DIR}",
    #     stdout=asyncio.subprocess.PIPE,  # We still need these to create the process
    #     stderr=asyncio.subprocess.PIPE,  # Even if we don't use the output.
    # )
    # stdout, stderr = await process.communicate() # Get the output
    # if process.returncode != 0:
    #     print(f"Failed to copy to {host}. Command exited with error code {process.returncode}")
    #     print(stderr.decode())
    # return process.returncode  # Return the exit code

async def main():
    for host in REMOTE_HOSTS:
        asyncio.create_task(copy_up(host))

asyncio.run(main())


import asyncio
import argparse
import time
import math
import random
import logging
import websockets
from dataclasses import dataclass

# Importing generated protobufs
# Ensure these are in your python path
from nf_robot.generated.nf import telemetry, common, control

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger("SimRobot")

# Constants
ROOM_SIZE_X = 5.0
ROOM_SIZE_Y = 5.0
HALF_ROOM_X = ROOM_SIZE_X / 2.0
HALF_ROOM_Y = ROOM_SIZE_Y / 2.0
ANCHOR_HEIGHT = 2.5
GRIPPER_OFFSET_Z = 0.53
UPDATE_RATE_HZ = 30
DT = 1.0 / UPDATE_RATE_HZ
ROBOT_ID = 'simulated_robot_1'

@dataclass
class RobotState:
    # Gantry kinematics (Origin is now center of room, z=0 is floor)
    pos_x: float = 0.0
    pos_y: float = 0.0
    pos_z: float = 1.0
    vel_x: float = 0.0
    vel_y: float = 0.0
    vel_z: float = 0.0
    
    # Gripper state
    wrist_angle: float = 0.0
    finger_angle: float = 0.0 # -90 to 90
    
    # Simulation internals
    target_vel_x: float = 0.0
    target_vel_y: float = 0.0
    target_vel_z: float = 0.0
    
    last_update: float = 0.0

def euler_to_rodrigues(roll, pitch, yaw):
    """
    Convert Euler angles (radians) to Rodrigues rotation vector (Vec3).
    Rodrigues vector r: direction is axis of rotation, magnitude is angle in radians.
    """
    # 1. Convert Euler to Quaternion (w, x, y, z)
    cy = math.cos(yaw * 0.5)
    sy = math.sin(yaw * 0.5)
    cp = math.cos(pitch * 0.5)
    sp = math.sin(pitch * 0.5)
    cr = math.cos(roll * 0.5)
    sr = math.sin(roll * 0.5)

    w = cr * cp * cy + sr * sp * sy
    x = sr * cp * cy - cr * sp * sy
    y = cr * sp * cy + sr * cp * sy
    z = cr * cp * sy - sr * sp * cy

    # 2. Convert Quaternion to Axis-Angle (Rodrigues)
    # Angle theta = 2 * acos(w)
    # Axis v = (x, y, z) / sin(theta/2)
    # Rodrigues r = theta * v
    
    # Calculate magnitude of the vector part (sin(theta/2))
    sin_half_theta_sq = x*x + y*y + z*z
    
    # Avoid division by zero for very small rotations
    if sin_half_theta_sq < 1e-7:
        return common.Vec3(x=0.0, y=0.0, z=0.0)
    
    sin_half_theta = math.sqrt(sin_half_theta_sq)
    
    # Use atan2 for stable angle calculation
    theta = 2.0 * math.atan2(sin_half_theta, w)
    
    scale = theta / sin_half_theta
    
    return common.Vec3(x=x*scale, y=y*scale, z=z*scale)

def get_anchor_poses():
    """
    Define 4 anchors in corners of a 5m square room, 2.5m high.
    Origin is center of room (0,0).
    Rotated along Z to look outwards (previous inward angle + 180).
    """
    poses = []
    
    # Corner definitions: (x, y) -> facing angle (degrees)
    # Forward is +Y (0 deg).
    # Previous angles: -45, 45, 135, 225.
    # New angles: +180 to each.
    
    corners = [
        (-2.5, -2.5, -45 + 180), # Anchor 0: Bottom-Left
        (2.5, -2.5, 45 + 180),   # Anchor 1: Bottom-Right
        (2.5, 2.5, 135 + 180),   # Anchor 2: Top-Right
        (-2.5, 2.5, 225 + 180)   # Anchor 3: Top-Left
    ]

    for x, y, deg in corners:
        # Convert degrees to radians for yaw
        yaw_rad = math.radians(deg)
        
        # Assuming Z-up coordinate system where rotation is around Z
        rotation = euler_to_rodrigues(0, 0, yaw_rad)
        
        pos = common.Vec3(x=x, y=y, z=ANCHOR_HEIGHT)
        poses.append(common.Pose(position=pos, rotation=rotation))
        
    return telemetry.AnchorPoses(poses=poses)

async def simulate_component_connection(websocket, is_gripper, anchor_num=0):
    """
    Simulates the connection lifecycle of a component:
    Connecting (2s) -> Connected (Websocket) -> Connected (Video).
    """
    # Phase 1: Connecting (2 seconds)
    status_msg = telemetry.ComponentConnStatus(
        is_gripper=is_gripper,
        anchor_num=anchor_num if not is_gripper else 0,
        websocket_status=telemetry.ConnStatus.CONNECTING,
        video_status=telemetry.ConnStatus.NOT_DETECTED,
        ip_address="192.168.1.10" + str(anchor_num if not is_gripper else 9),
        gripper_model=telemetry.GripperModel.PILOT if is_gripper else None
    )
    
    update = telemetry.TelemetryBatchUpdate(
        robot_id=ROBOT_ID,
        updates=[telemetry.TelemetryItem(
            component_conn_status=status_msg,
            retain_key=f"conn_status_{'gripper' if is_gripper else f'anchor_{anchor_num}'}"
        )]
    )
    
    try:
        await websocket.send(bytes(update))
    except Exception as e:
        logger.error(f"Failed to send connecting status: {e}")
        return

    await asyncio.sleep(3.0)

    # Phase 2: Websocket Connected
    status_msg.websocket_status = telemetry.ConnStatus.CONNECTED
    update.updates[0].component_conn_status = status_msg
    
    try:
        await websocket.send(bytes(update))
    except Exception as e:
        logger.error(f"Failed to send connected status: {e}")
        return
        
    await asyncio.sleep(3.0) # visible delay before video comes online

    # Phase 3: Video Connected
    status_msg.video_status = telemetry.ConnStatus.CONNECTED
    update.updates[0].component_conn_status = status_msg
    
    try:
        await websocket.send(bytes(update))
    except Exception as e:
        logger.error(f"Failed to send video status: {e}")

async def physics_loop(websocket, state: RobotState):
    """
    Main loop producing 30fps telemetry updates.
    """
    state.last_update = time.time()
    
    while True:
        now = time.time()
        dt_actual = now - state.last_update
        state.last_update = now

        # Update Gantry Position
        # Simple Euler integration
        state.vel_x += (state.target_vel_x - state.vel_x) * 0.1 # Simple smoothing
        state.vel_y += (state.target_vel_y - state.vel_y) * 0.1
        state.vel_z += (state.target_vel_z - state.vel_z) * 0.1

        state.pos_x += state.vel_x * dt_actual
        state.pos_y += state.vel_y * dt_actual
        state.pos_z += state.vel_z * dt_actual

        # Clamp to room boundaries (Origin at center, so +/- half size)
        state.pos_x = max(-HALF_ROOM_X, min(HALF_ROOM_X, state.pos_x))
        state.pos_y = max(-HALF_ROOM_Y, min(HALF_ROOM_Y, state.pos_y))
        state.pos_z = max(0.0, min(ANCHOR_HEIGHT, state.pos_z))

        # --- Prepare Telemetry ---
        
        # 1. Position Estimate
        # Gripper is 53cm below gantry
        gripper_pos = common.Vec3(x=state.pos_x, y=state.pos_y, z=state.pos_z - GRIPPER_OFFSET_Z)
        
        # Random rotation for gripper to simulate "guessing" if we don't have commanded data
        # Gripper rotated -90 degrees around Y (Pitch = -90)
        gripper_rot = euler_to_rodrigues(0, 0, math.radians(state.wrist_angle))
        
        pos_est = telemetry.PositionEstimate(
            gantry_position=common.Vec3(x=state.pos_x, y=state.pos_y, z=state.pos_z),
            gantry_velocity=common.Vec3(x=state.vel_x, y=state.vel_y, z=state.vel_z),
            gripper_pose=common.Pose(position=gripper_pos, rotation=gripper_rot),
            data_ts=now,
            slack=[False, False, False, False]
        )

        # 2. Position Factors
        # Add noise to "real" position for visual
        noise_level = 0.02
        vis_pos = common.Vec3(
            x=state.pos_x + random.uniform(-noise_level, noise_level),
            y=state.pos_y + random.uniform(-noise_level, noise_level),
            z=state.pos_z + random.uniform(-noise_level, noise_level)
        )
        pos_factors = telemetry.PositionFactors(
            visual_pos=vis_pos,
            visual_vel=common.Vec3(x=state.vel_x, y=state.vel_y, z=state.vel_z),
            hanging_pos=common.Vec3(x=state.pos_x, y=state.pos_y, z=state.pos_z), # Ideal hanging
            hanging_vel=common.Vec3(x=state.vel_x, y=state.vel_y, z=state.vel_z)
        )

        # 3. Gripper Sensors
        # Calculate real range to floor (z=0)
        range_to_floor = max(0.0, (state.pos_z - GRIPPER_OFFSET_Z))
        
        # Simulate pressure only if close to floor and gripper is closed
        simulated_pressure = 0.0
        if state.finger_angle > 45 and range_to_floor < 0.1:
            simulated_pressure = random.uniform(0.5, 1.5)

        grip_sensors = telemetry.GripperSensors(
            range=range_to_floor + random.uniform(-0.005, 0.005), # Range from palm to floor
            angle=state.finger_angle,
            pressure=simulated_pressure, 
            wrist=state.wrist_angle
        )

        # 4. Commanded Velocity
        # The velocity we are actually trying to achieve (target_vel)
        cmd_vel = telemetry.CommandedVelocity(
            velocity=common.Vec3(x=state.target_vel_x, y=state.target_vel_y, z=state.target_vel_z)
        )

        # Construct Batch Update
        batch = telemetry.TelemetryBatchUpdate(
            robot_id=ROBOT_ID,
            updates=[
                telemetry.TelemetryItem(pos_estimate=pos_est, retain_key="pos_estimate"),
                telemetry.TelemetryItem(pos_factors_debug=pos_factors),
                telemetry.TelemetryItem(grip_sensors=grip_sensors, retain_key="grip_sensors"),
                telemetry.TelemetryItem(last_commanded_vel=cmd_vel, retain_key="cmd_vel")
            ]
        )

        try:
            await websocket.send(bytes(batch))
        except Exception as e:
            logger.error(f"Send failed in physics loop: {e}")
            break

        await asyncio.sleep(DT)

async def receive_loop(websocket, state: RobotState):
    """
    Listens for ControlBatchUpdate messages and updates state.
    """
    async for message in websocket:
        # Letting exceptions propagate if parsing fails
        try:
            batch = control.ControlBatchUpdate().parse(message)
            
            for item in batch.updates:
                # betterproto2 exposes oneofs as attributes.
                # We check the ones we care about.
                
                if item.command:
                    cmd = item.command
                    logger.info(f"Received CommonCommand: {cmd.name}")
                    if cmd.name == control.Command.STOP_ALL:
                        logger.info("STOPPING ALL MOTION")
                        state.target_vel_x = 0
                        state.target_vel_y = 0
                        state.target_vel_z = 0
                        state.vel_x = 0
                        state.vel_y = 0
                        state.vel_z = 0

                elif item.move:
                    move = item.move
                    logger.info(f"Received Move: {move}")
                    
                    # --- Velocity Update Logic ---
                    # We update velocity if either direction or speed is provided.
                    # This handles:
                    # 1. CombinedMove(direction=Vec3(), speed=0.0) -> Stop
                    # 2. CombinedMove(speed=0.0) -> Stop
                    # 3. CombinedMove(direction=Vec3(1,0,0), speed=1.0) -> Move
                    # 4. CombinedMove(direction=Vec3(1,0,0)) -> Move (direction is velocity)
                    
                    should_update_vel = False
                    new_vel_x, new_vel_y, new_vel_z = 0.0, 0.0, 0.0

                    # Check for explicit stop command (speed=0)
                    if move.speed is not None and move.speed == 0.0:
                        should_update_vel = True
                        # Velocities remain 0.0
                        
                    elif move.direction is not None:
                        should_update_vel = True
                        mag = math.sqrt(move.direction.x**2 + move.direction.y**2 + move.direction.z**2)
                        
                        if mag > 0:
                            if move.speed is not None:
                                # Speed provided: normalize direction and scale
                                scale = move.speed / mag
                                new_vel_x = move.direction.x * scale
                                new_vel_y = move.direction.y * scale
                                new_vel_z = move.direction.z * scale
                            else:
                                # Speed not provided: direction is velocity
                                new_vel_x = move.direction.x
                                new_vel_y = move.direction.y
                                new_vel_z = move.direction.z
                        else:
                            # Direction is (0,0,0) -> Stop
                            # Velocities remain 0.0
                            pass

                    if should_update_vel:
                        state.target_vel_x = new_vel_x
                        state.target_vel_y = new_vel_y
                        state.target_vel_z = new_vel_z
                    
                    # Update Finger
                    if move.finger is not None:
                        state.finger_angle = move.finger
                        
                    # Update Wrist
                    if move.wrist is not None:
                        state.wrist_angle = move.wrist

                else:
                    # Log other commands but do nothing
                    logger.info(f"Ignored control command item: {item}")
        except Exception as e:
            logger.error(f"Error handling message: {e}")
            import traceback
            traceback.print_exc()

async def main():
    parser = argparse.ArgumentParser(description="Simulated Cable Robot")
    parser.add_argument("server_address", help="WebSocket server address (e.g., ws://localhost:8080)")
    args = parser.parse_args()

    uri = f'{args.server_address}/telemetry/{ROBOT_ID}'
    logger.info(f"Connecting to {uri}...")

    async with websockets.connect(uri) as websocket:
        logger.info("Connected to server.")
        
        state = RobotState()
        
        # 1. Send Anchor Poses immediately
        anchor_poses_msg = get_anchor_poses()
        init_update = telemetry.TelemetryBatchUpdate(
            robot_id=ROBOT_ID,
            updates=[telemetry.TelemetryItem(new_anchor_poses=anchor_poses_msg, retain_key="anchor_poses")]
        )
        await websocket.send(bytes(init_update))
        logger.info("Sent initial AnchorPoses.")

        # 2. Start Connection Simulation Tasks (Background)
        conn_tasks = []
        # 4 Anchors
        for i in range(4):
            await asyncio.sleep(1.0)
            conn_tasks.append(asyncio.create_task(simulate_component_connection(websocket, is_gripper=False, anchor_num=i)))
        # 1 Gripper
        conn_tasks.append(asyncio.create_task(simulate_component_connection(websocket, is_gripper=True)))

        # 3. Start Loops
        physics_task = asyncio.create_task(physics_loop(websocket, state))
        receive_task = asyncio.create_task(receive_loop(websocket, state))

        # Wait for loops (they generally run forever until connection drop)
        await asyncio.gather(physics_task, receive_task, *conn_tasks)

if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        logger.info("Simulation stopped by user.")


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import Dataset, DataLoader
import cv2
import os
import json
import numpy as np
import argparse
import random
import uuid
import shutil
import subprocess
from importlib.resources import files

# ==========================================
# CONFIGURATION
# ==========================================
DEFAULT_REPO_ID = "naavox/target-heatmap-dataset"
DEFAULT_MODEL_PATH = "models/target_heatmap.pth"
LOCAL_DATASET_ROOT = "target_heatmap_data"
HEATMAP_UNPROCESSED_DIR = "target_heatmap_data_unlabeled"
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"

# Network Input Resolution
# 960x544 is divisible by 32 (standard for CNNs), ensuring perfect alignment 
# through pooling and upsampling layers without rounding errors.
HM_IMAGE_RES = (960, 544) 

# Labeling Source Resolution
SOURCE_RES = (1920, 1080)

MINIMUM_CONFIDENCE = 0.95 # during eval

# ==========================================
# MODEL DEFINITION
# ==========================================

class TargetHeatmapNet(nn.Module):
    """
    Learns a heatmap from images that have one or more labeled points.
    Input images are expected to be 960x544.
    """

    def __init__(self):
        super().__init__()
        # Encoder
        self.enc1 = self.conv_block(3, 32)
        self.pool1 = nn.MaxPool2d(2)
        self.enc2 = self.conv_block(32, 64)
        self.pool2 = nn.MaxPool2d(2)
        self.enc3 = self.conv_block(64, 128)
        self.pool3 = nn.MaxPool2d(2)
        
        # Bottleneck
        self.bottleneck = self.conv_block(128, 256)
        
        # Decoder
        # Since input is divisible by 8 (2^3), we can use standard fixed Upsample layers
        self.up3 = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
        self.dec3 = self.conv_block(256 + 128, 128)
        self.up2 = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
        self.dec2 = self.conv_block(128 + 64, 64)
        self.up1 = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
        self.dec1 = self.conv_block(64 + 32, 32)
        self.final = nn.Conv2d(32, 1, kernel_size=1)

    def conv_block(self, in_c, out_c):
        return nn.Sequential(
            nn.Conv2d(in_c, out_c, 3, padding=1),
            nn.BatchNorm2d(out_c),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_c, out_c, 3, padding=1),
            nn.BatchNorm2d(out_c),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        e1 = self.enc1(x)
        e2 = self.enc2(self.pool1(e1))
        e3 = self.enc3(self.pool2(e2))
        
        b = self.bottleneck(self.pool3(e3))
        
        # Dimensions align perfectly with 960x544 input
        d3 = self.dec3(torch.cat([self.up3(b), e3], dim=1))
        d2 = self.dec2(torch.cat([self.up2(d3), e2], dim=1))
        d1 = self.dec1(torch.cat([self.up1(d2), e1], dim=1))
        
        return torch.sigmoid(self.final(d1))

# ==========================================
# DATASET & UTILS
# ==========================================

def generate_blob(x_grid, y_grid, cx, cy, sigma=15):
    """Generates a Gaussian blob at (cx, cy)."""
    return np.exp(-((x_grid - cx)**2 + (y_grid - cy)**2) / (2 * sigma**2))

class DobbyDataset(Dataset):
    def __init__(self, root_dir):
        self.data_dir = os.path.join(root_dir, "train")
        self.metadata_path = os.path.join(self.data_dir, "metadata.jsonl")
        
        if not os.path.exists(self.metadata_path):
            raise FileNotFoundError(f"Could not find metadata at {self.metadata_path}")
            
        self.samples = []
        with open(self.metadata_path, 'r') as f:
            for line in f:
                if line.strip():
                    self.samples.append(json.loads(line))

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        item = self.samples[idx]
        img_path = os.path.join(self.data_dir, item["file_name"])
        
        img = cv2.imread(img_path)
        if img is None:
            raise ValueError(f"Failed to load image: {img_path}")
            
        h, w = img.shape[:2]
        img_tensor = torch.from_numpy(img).permute(2, 0, 1).float() / 255.0

        x_grid = np.arange(0, w, 1, float)
        y_grid = np.arange(0, h, 1, float)[:, np.newaxis]
        
        combined_heatmap = np.zeros((h, w), dtype=np.float32)
        
        for pt in item.get("points", []):
            if isinstance(pt, (list, tuple)):
                cx, cy = pt[0], pt[1]
            elif isinstance(pt, dict):
                cx, cy = pt['x'], pt['y']
            else:
                continue
                
            combined_heatmap = np.maximum(combined_heatmap, generate_blob(x_grid, y_grid, cx, cy))
            
        return img_tensor, torch.from_numpy(combined_heatmap).float().unsqueeze(0)

def extract_targets_from_heatmap(heatmap: np.ndarray, top_n: int = 10, threshold: float = 0.5):
    """
    Extracts the centers of high-confidence blobs from a heatmap.
    Returns sorted list of (norm_x, norm_y, confidence).
    """
    mask = (heatmap > threshold).astype(np.uint8) * 255

    # RETR_EXTERNAL to ignore holes inside blobs
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    candidates = []
    for contour in contours:
        x, y, w, h = cv2.boundingRect(contour)
        
        roi = heatmap[y:y+h, x:x+w]
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(roi)
        
        global_x = x + max_loc[0]
        global_y = y + max_loc[1]
        
        candidates.append((global_x, global_y, max_val))

    candidates.sort(key=lambda k: k[2], reverse=True)

    height, width = heatmap.shape
    results = []
    
    for c in candidates[:top_n]:
        norm_x = c[0] / width
        norm_y = c[1] / height
        confidence = c[2]
        if confidence > MINIMUM_CONFIDENCE:
            results.append((norm_x, norm_y, confidence))

    return np.array(results)

# ==========================================
# TRAINING LOOP
# ==========================================

def train(args):
    from huggingface_hub import snapshot_download
    print(f"Downloading/Loading dataset from {args.dataset_id}...")
    dataset_path = snapshot_download(repo_id=args.dataset_id, repo_type="dataset")
    print(f"Dataset available at: {dataset_path}")

    dataset = DobbyDataset(dataset_path)
    dataloader = DataLoader(dataset, batch_size=args.batch_size, shuffle=True)
    
    os.makedirs(os.path.dirname(args.model_path), exist_ok=True)
    model = TargetHeatmapNet().to(DEVICE)
    optimizer = optim.Adam(model.parameters(), lr=args.lr)
    
    criterion = nn.BCEWithLogitsLoss() 

    print(f"Starting training on {len(dataset)} images for {args.epochs} epochs...")
    print(f"Device: {DEVICE}")

    for epoch in range(args.epochs):
        total_loss = 0
        model.train()
        
        for imgs, maps in dataloader:
            imgs, maps = imgs.to(DEVICE), maps.to(DEVICE)
            
            optimizer.zero_grad()
            preds = model(imgs)
            loss = criterion(preds, maps)
            loss.backward()
            optimizer.step()
            
            total_loss += loss.item()
            
        print(f"Epoch {epoch+1}/{args.epochs}, Loss: {total_loss / len(dataloader):.5f}")

    torch.save(model.state_dict(), args.model_path)
    print(f"Model saved to {args.model_path}")

# ==========================================
# EVALUATION TOOL
# ==========================================

def run_inference(model, img_bgr):
    """
    Helper to run model on a single BGR image and return overlay.
    """
    img_tensor = torch.from_numpy(img_bgr).permute(2, 0, 1).float() / 255.0
    batch = img_tensor.unsqueeze(0).to(DEVICE)

    with torch.no_grad():
        heatmap_out = model(batch)
    heatmap_np = heatmap_out.squeeze().cpu().numpy()
    
    # img_bgr is used directly for background (no channel swap needed)
    img_display = img_bgr.copy()
    
    heatmap_vis = (heatmap_np * 255).astype(np.uint8)
    heatmap_color = cv2.applyColorMap(heatmap_vis, cv2.COLORMAP_JET)
    
    overlay = cv2.addWeighted(img_display, 0.8, heatmap_color, 0.4, 0)

    targets = extract_targets_from_heatmap(heatmap_np)
    
    for x, y, confidence in targets:
        x = int(x * HM_IMAGE_RES[0])
        y = int(y * HM_IMAGE_RES[1])
        
        box_size = 20
        top_left =     (x - box_size, y - box_size)
        bottom_right = (x + box_size, y + box_size)
        cv2.rectangle(overlay, top_left, bottom_right, (0, 255, 0), 2)
        conf_text = f"{confidence:.2f}"
        cv2.putText(overlay, conf_text, (x - 10, y - 25), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 1)
        
    return overlay

def eval_mode(args):
    from huggingface_hub import snapshot_download
    print(f"Loading model from {args.model_path}...")
    model = TargetHeatmapNet().to(DEVICE)
    model.load_state_dict(torch.load(args.model_path, map_location=DEVICE))
    model.eval()

    window_name = "Target Heatmap"
    cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
    
    # Mode 1: Video Stream
    if args.uri:
        source = args.uri
        # Convert to int if user passed a webcam index (e.g. "0")
        if source.isdigit():
            source = int(source)
            
        print(f"Opening video source: {source}")
        cap = cv2.VideoCapture(source)
        
        if not cap.isOpened():
            print(f"Error: Could not open video source {source}")
            return

        fps = cap.get(cv2.CAP_PROP_FPS)
        if fps is None or fps <= 0 or np.isnan(fps):
            fps = 30.0
        print(f"Stream FPS: {fps}")

        cv2.resizeWindow(window_name, HM_IMAGE_RES[0], HM_IMAGE_RES[1])
        print("Controls: [Q] Quit")

        # FFMPEG Recording Setup
        recorder = None
        if args.record:
            # We are writing raw BGR24 frames to stdin
            command = [
                'ffmpeg',
                '-y',                    # Overwrite output
                '-f', 'rawvideo',        # Input format
                '-vcodec', 'rawvideo',
                '-s', f'{HM_IMAGE_RES[0]}x{HM_IMAGE_RES[1]}', # Size
                '-pix_fmt', 'bgr24',     # OpenCV uses BGR
                '-r', str(fps),          # Input framerate
                '-i', '-',               # Read from stdin
                '-c:v', 'libx264',       # Output codec
                '-pix_fmt', 'yuv420p',   # Pixel format for compatibility
                '-preset', 'fast',       # Encoding speed
                args.record              # Output filename
            ]
            print(f"Starting recording: {' '.join(command)}")
            try:
                recorder = subprocess.Popen(command, stdin=subprocess.PIPE)
            except FileNotFoundError:
                print("Error: ffmpeg not found. Please install ffmpeg to record.")
                return

        while True:
            ret, frame = cap.read()
            if not ret:
                print("End of stream.")
                break
            
            # Ensure frame matches network input size
            frame_resized = cv2.resize(frame, HM_IMAGE_RES)
            
            overlay = run_inference(model, frame_resized)
            
            # Write to ffmpeg stdin
            if recorder:
                try:
                    recorder.stdin.write(overlay.tobytes())
                except BrokenPipeError:
                    print("FFmpeg recording stopped unexpectedly.")
                    recorder = None

            cv2.imshow(window_name, overlay)
            
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break
        
        cap.release()
        if recorder:
            recorder.stdin.close()
            recorder.wait()
            print(f"Saved recording to {args.record}")
        
    # Mode 2: Dataset Evaluation
    else:
        print(f"Downloading dataset {args.dataset_id} for samples...")
        dataset_path = snapshot_download(repo_id=args.dataset_id, repo_type="dataset")
        
        data_dir = os.path.join(dataset_path, "eval")
        metadata_path = os.path.join(data_dir, "metadata.jsonl")
        
        if not os.path.exists(metadata_path):
            print(f"No eval metadata found at {metadata_path}. Did you run 'split' on the dataset?")
            return

        samples = []
        with open(metadata_path, 'r') as f:
            for line in f:
                if line.strip(): samples.append(json.loads(line))
                
        print(f"Loaded {len(samples)} evaluation samples.")
        print("Controls: [SPACE] Next, [Q] Quit")
        
        cv2.resizeWindow(window_name, HM_IMAGE_RES[0] * 2, HM_IMAGE_RES[1] * 2)

        while True:
            sample = random.choice(samples)
            img_path = os.path.join(data_dir, sample["file_name"])
            img_input = cv2.imread(img_path)
            
            if img_input is None: 
                continue

            overlay = run_inference(model, img_input)
            
            cv2.imshow(window_name, overlay)
            if cv2.waitKey(0) & 0xFF == ord('q'):
                break
                
    cv2.destroyAllWindows()

# ==========================================
# LABELING TOOL
# ==========================================

# State
current_clicks = []
current_image = None
current_image_path = None

def mouse_callback(event, x, y, flags, param):
    global current_clicks
    if event == cv2.EVENT_LBUTTONDOWN:
        current_clicks.append((x, y))
        print(f"Added Point (Source Res): {x}, {y}")
        draw_interface()
    elif event == cv2.EVENT_RBUTTONDOWN:
        if current_clicks:
            removed = current_clicks.pop()
            print(f"Removed Point: {removed}")
            draw_interface()

def draw_interface():
    global current_image, current_clicks
    if current_image is None:
        return

    display = current_image.copy()

    for pt in current_clicks:
        cv2.circle(display, pt, 5, (0, 255, 0), -1)

    status_text = f"Points: {len(current_clicks)} | Res: {SOURCE_RES}"
    cv2.putText(display, status_text, (10, 30), 
                cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 255, 255), 2)

    cv2.imshow("Labeler", display)

def label_mode(args):
    global current_clicks, current_image, current_image_path

    TRAIN_DIR = os.path.join(LOCAL_DATASET_ROOT, "train")
    METADATA_PATH = os.path.join(TRAIN_DIR, "metadata.jsonl")
    
    if not os.path.exists(TRAIN_DIR): os.makedirs(TRAIN_DIR)
    
    readme_path = os.path.join(LOCAL_DATASET_ROOT, "README.md")
    if not os.path.exists(readme_path):
        with open(readme_path, "w") as f:
            f.write("---\nconfigs:\n- config_name: default\n  data_files:\n  - split: train\n    path: train/metadata.jsonl\n---\n")

    cv2.namedWindow("Labeler", cv2.WINDOW_NORMAL)
    cv2.resizeWindow("Labeler", SOURCE_RES[0], SOURCE_RES[1])
    cv2.setMouseCallback("Labeler", mouse_callback)
    
    print("Labeler Started.")
    print("Files are loaded from:", HEATMAP_UNPROCESSED_DIR)
    print(f"Expected Input Res: {SOURCE_RES}")
    print(f"Saved Target Res:   {HM_IMAGE_RES}")
    print("\n--- Instructions ---")
    print("1. Left-Click:   Mark a spot.")
    print("2. Right-Click:  Undo last mark.")
    print("3. SPACE:        Save (resizes img & points) and Next.")
    print("4. 'n':          Skip to next random frame.")
    print("5. 'q':          Quit and ask to upload to Hugging Face.")

    while True:
        if not os.path.exists(HEATMAP_UNPROCESSED_DIR):
            print(f"No unprocessed directory: {HEATMAP_UNPROCESSED_DIR}")
            break
            
        files = [f for f in os.listdir(HEATMAP_UNPROCESSED_DIR) if f.endswith('.jpg')]
        if not files:
            print("No more files to process.")
            upload_prompt(args)
            break
            
        fn = random.choice(files)
        current_image_path = os.path.join(HEATMAP_UNPROCESSED_DIR, fn)
    
        current_image = cv2.imread(current_image_path)
        if current_image is None:
            continue
            
        # Ensure we are working with expected source resolution or warn/resize
        h, w = current_image.shape[:2]
        if (w, h) != SOURCE_RES:
            print(f"Warning: Image {fn} is {w}x{h}, expected {SOURCE_RES}. Visuals might be skewed.")

        current_clicks = [] 
        
        draw_interface()

        save_it = False
        while True:
            key = cv2.waitKey(1) & 0xFF

            if key == ord('q'):
                cv2.destroyAllWindows()
                upload_prompt(args)
                return
                
            elif key == ord('n'):
                print("Skipped frame.")
                break 
                
            elif key == ord(' '):
                save_it = True
                break

        if save_it:
            # Transform points from SOURCE_RES to HM_IMAGE_RES
            scale_x = HM_IMAGE_RES[0] / SOURCE_RES[0]
            scale_y = HM_IMAGE_RES[1] / SOURCE_RES[1]
            
            scaled_points = []
            for px, py in current_clicks:
                scaled_points.append((int(px * scale_x), int(py * scale_y)))

            # Resize image
            resized_img = cv2.resize(current_image, HM_IMAGE_RES, interpolation=cv2.INTER_AREA)

            # Generate filename and save
            new_id = str(uuid.uuid4())
            new_fn = f"{new_id}.jpg"
            new_path = os.path.join(TRAIN_DIR, new_fn)
            
            cv2.imwrite(new_path, resized_img)
            
            # Remove original from unprocessed
            os.remove(current_image_path)
            
            # Write Metadata
            entry = {
                "file_name": new_fn,
                "points": scaled_points
            }
            with open(METADATA_PATH, 'a') as f:
                f.write(json.dumps(entry) + "\n")
            
            print(f"Saved {len(scaled_points)} points -> {new_fn} (Resized to {HM_IMAGE_RES})")

def upload_prompt(args):
    from huggingface_hub import HfApi, create_repo
    if not os.path.exists(LOCAL_DATASET_ROOT): return
    
    print("\n" + "="*30)
    print(f"Data organized in '{LOCAL_DATASET_ROOT}'")
    confirm = input(f"Upload to {args.dataset_id}? (y/n): ").strip().lower()
    
    if confirm == 'y':
        api = HfApi()
        create_repo(args.dataset_id, repo_type="dataset", exist_ok=True)
        api.upload_folder(
            folder_path=LOCAL_DATASET_ROOT,
            repo_id=args.dataset_id,
            repo_type="dataset"
        )
        print("Uploaded successfully.")

def split_and_upload(args):
    print(f"Preparing to split dataset in {LOCAL_DATASET_ROOT}...")
    
    train_dir = os.path.join(LOCAL_DATASET_ROOT, "train")
    eval_dir = os.path.join(LOCAL_DATASET_ROOT, "eval")
    train_meta = os.path.join(train_dir, "metadata.jsonl")
    eval_meta = os.path.join(eval_dir, "metadata.jsonl")

    if not os.path.exists(train_dir): os.makedirs(train_dir)
    if not os.path.exists(eval_dir): os.makedirs(eval_dir)

    all_samples = []

    def load_samples(meta_path, source_split):
        if os.path.exists(meta_path):
            with open(meta_path, 'r') as f:
                for line in f:
                    if line.strip():
                        entry = json.loads(line)
                        entry['_current_split'] = source_split
                        all_samples.append(entry)

    load_samples(train_meta, "train")
    load_samples(eval_meta, "eval")

    if not all_samples:
        print("No data found in local folders.")
        return

    print(f"Found {len(all_samples)} total samples. Shuffling and splitting 90/10...")
    random.shuffle(all_samples)

    split_idx = int(len(all_samples) * 0.9)
    train_set = all_samples[:split_idx]
    eval_set = all_samples[split_idx:]
    
    print(f"New distribution -> Train: {len(train_set)} | Eval: {len(eval_set)}")

    def process_split(sample_list, target_split, target_dir, target_meta_path):
        with open(target_meta_path, 'w') as f:
            for entry in sample_list:
                current_split = entry.pop('_current_split')
                fname = entry['file_name']
                
                if current_split != target_split:
                    src_path = os.path.join(LOCAL_DATASET_ROOT, current_split, fname)
                    dst_path = os.path.join(target_dir, fname)
                    
                    if os.path.exists(src_path):
                        shutil.move(src_path, dst_path)
                    else:
                        print(f"Warning: File missing at {src_path}")
                
                f.write(json.dumps(entry) + "\n")

    process_split(train_set, "train", train_dir, train_meta)
    process_split(eval_set, "eval", eval_dir, eval_meta)

    readme_path = os.path.join(LOCAL_DATASET_ROOT, "README.md")
    with open(readme_path, "w") as f:
        f.write("---\nconfigs:\n- config_name: default\n  data_files:\n  - split: train\n    path: train/metadata.jsonl\n  - split: test\n    path: eval/metadata.jsonl\n---\n")

    upload_prompt(args)

# ==========================================
# MAIN ENTRY POINT
# ==========================================

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Target Heatmap ML Tools")
    subparsers = parser.add_subparsers(dest="command", required=True)

    # Train Command
    train_parser = subparsers.add_parser("train")
    train_parser.add_argument("--dataset_id", type=str, default=DEFAULT_REPO_ID)
    train_parser.add_argument("--model_path", type=str, default=DEFAULT_MODEL_PATH)
    train_parser.add_argument("--epochs", type=int, default=200)
    train_parser.add_argument("--batch_size", type=int, default=10)
    train_parser.add_argument("--lr", type=float, default=1e-4)

    # Eval Command
    eval_parser = subparsers.add_parser("eval")
    eval_parser.add_argument("--dataset_id", type=str, default=DEFAULT_REPO_ID)
    eval_parser.add_argument("--model_path", type=str, default=DEFAULT_MODEL_PATH)
    eval_parser.add_argument("--uri", type=str, default=None, help="Video file path or camera index")
    eval_parser.add_argument("--record", type=str, default=None, help="Path to save MP4 recording (only works with --uri)")


    # Label Command
    label_parser = subparsers.add_parser("label")
    label_parser.add_argument("--dataset_id", type=str, default=DEFAULT_REPO_ID)

    # Split Command
    split_parser = subparsers.add_parser("split")
    split_parser.add_argument("--dataset_id", type=str, default=DEFAULT_REPO_ID)

    args = parser.parse_args()

    if args.command == "train":
        train(args)
    elif args.command == "eval":
        eval_mode(args)
    elif args.command == "label":
        label_mode(args)
    elif args.command == "split":
        split_and_upload(args)